# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
from jax import random
from jax import tree_util
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import ScipyBoundedLeastSquares
from nemos.third_party.jaxopt.jaxopt import ScipyBoundedMinimize
from nemos.third_party.jaxopt.jaxopt import ScipyLeastSquares
from nemos.third_party.jaxopt.jaxopt import ScipyMinimize
from nemos.third_party.jaxopt.jaxopt import ScipyRootFinding
from nemos.third_party.jaxopt.jaxopt._src import scipy_wrappers
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_scalar_mul

import numpy as onp
import scipy as osp
from sklearn import datasets


class JnpToOnpTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()

    m, n = 10, 9
    self.x_split_indices = [4, 6, 9]
    self.x_reshape_shape = [-1, 2]
    self.y_split_indices = [6, 8]
    self.y_reshape_shape = [-1, 3]

    key = random.PRNGKey(0)
    key_x, key_w1, key_w2 = random.split(key, 3)
    self.x = random.normal(key_x, [m])
    self.w1 = random.normal(key_w1, [m, m])
    self.w2 = random.normal(key_w2, [m, n])

  @staticmethod
  def _do_pytree(array, split_indices, reshape_shape):
    """Creates a (hardcoded) PyTree out of flat 1D array for testing purposes."""
    pytree = jnp.split(array, split_indices)
    pytree[0] = pytree[0].reshape(reshape_shape)
    pytree[-1] = {'-1': pytree[-1]}
    return pytree

  @staticmethod
  def _undo_pytree(pytree):
    """Recovers flat 1D array from (hardcoded) PyTree for testing purposes."""
    pytree = pytree.copy()
    pytree[0] = pytree[0].reshape([-1])
    pytree[-1] = pytree[-1]['-1']
    return jnp.concatenate(pytree)

  def test_vals_and_jac(self):
    fun_flat = lambda x: jnp.dot(jnp.arctan(jnp.dot(x, self.w1)), self.w2)

    def fun(x_pytree):
      """Wraps fun_flat with (hardcoded) yTree input / output."""
      x = self._undo_pytree(x_pytree)
      y = fun_flat(x)
      return self._do_pytree(y, self.y_split_indices, self.y_reshape_shape)

    # Tests function output.
    x_pytree = self._do_pytree(self.x,
                               self.x_split_indices,
                               self.x_reshape_shape)
    y_pytree = fun(x_pytree)
    self.assertArraysAllClose(fun_flat(self.x), self._undo_pytree(y_pytree))

    # Tests jnp_to_onp.
    self.assertArraysAllClose(self._undo_pytree(y_pytree),
                              scipy_wrappers.jnp_to_onp(y_pytree))

    # Tests Jacobian.
    x_pytree_topology = scipy_wrappers.pytree_topology_from_example(x_pytree)
    y_pytree_topology = scipy_wrappers.pytree_topology_from_example(y_pytree)
    jac_jnp_to_onp = scipy_wrappers.make_jac_jnp_to_onp(x_pytree_topology,
                                                        y_pytree_topology)

    jac_flat = jax.jacrev(fun_flat)(self.x)
    jac_pytree = jax.jacrev(fun)(x_pytree)
    self.assertArraysAllClose(jac_flat, jac_jnp_to_onp(jac_pytree))


class ScipyMinimizeTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()

    self.n_samples, self.n_features, self.n_classes = 50, 5, 3
    self.data = datasets.make_classification(n_samples=self.n_samples,
                                             n_features=self.n_features,
                                             n_classes=self.n_classes,
                                             n_informative=3,
                                             random_state=0)
    self.default_l2reg = float(self.n_samples)

    self.solver_kwargs = {'method': 'L-BFGS-B',
                          'tol': 1e-5, 'maxiter': 500}

    def logreg_fun(params, *args, **kwargs):
      params = tree_util.tree_leaves(params)
      return objective.l2_multiclass_logreg(params[0], *args, **kwargs)
    self.logreg_fun = logreg_fun

    def logreg_with_intercept_fun(params, *args, **kwargs):
      return objective.l2_multiclass_logreg_with_intercept(
          (params['W'], params['b']), *args, **kwargs)
    self.logreg_with_intercept_fun = logreg_with_intercept_fun

  def test_maxiter_api(self):
    # tests that ScipyMinimize raises an exception when given contradictory arguments
    with self.assertRaises(ValueError):
      ScipyMinimize(fun=self.logreg_fun, maxiter=500, options={'maxiter': 100})

  @parameterized.product(method=[
    'Nelder-Mead',
    'Powell',
    'TNC',
  ])
  def test_no_njev(self, method):
    # test methods that do not return njev
    # see https://github.com/google/jaxopt/pull/542
    solver = ScipyMinimize(fun=self.logreg_fun, method=method)
    pytree_init = jnp.zeros([self.n_features, self.n_classes])
    solver.run(pytree_init, l2reg=self.default_l2reg, data=self.data)

  def test_callback(self):
    # test the callback API
    trace = []
    def cb(xk):
      # check that xk is a pytree with the right attributes
      self.assertIn('W', xk)
      self.assertIn('b', xk)
      trace.append(1)
    solver = ScipyMinimize(fun=self.logreg_with_intercept_fun, callback=cb)
    pytree_init = {'W': jnp.zeros([self.n_features, self.n_classes]),
                   'b': jnp.zeros([self.n_classes])}
    solver.run(pytree_init, l2reg=self.default_l2reg, data=self.data)

    # if the cb function has been called at least once, the trace list will have
    # at least one element
    self.assertNotEmpty(trace)

  def test_logreg(self):
    lbfgs = ScipyMinimize(fun=self.logreg_fun,
                          **self.solver_kwargs)

    pytree_init = jnp.zeros([self.n_features, self.n_classes])
    pytree_fit, _ = lbfgs.run(pytree_init,
                              l2reg=self.default_l2reg,
                              data=self.data)

    # Compare against sklearn.
    pytree_skl = test_util.logreg_skl(X=self.data[0],
                                      y=self.data[1],
                                      lam=self.default_l2reg,
                                      fit_intercept=False)
    for array_skl, array_fit in zip(tree_util.tree_leaves(pytree_skl),
                                    tree_util.tree_leaves(pytree_fit)):
      self.assertArraysAllClose(array_skl, array_fit, atol=1e-3)

  def test_logreg_with_intercept(self):
    lbfgs = ScipyMinimize(fun=self.logreg_with_intercept_fun,
                          **self.solver_kwargs)
    pytree_init = {'W': jnp.zeros([self.n_features, self.n_classes]),
                   'b': jnp.zeros([self.n_classes])}
    pytree_fit, _ = lbfgs.run(pytree_init,
                              l2reg=self.default_l2reg,
                              data=self.data)

    # Compare against sklearn.
    pytree_skl = test_util.logreg_skl(X=self.data[0],
                                      y=self.data[1],
                                      lam=self.default_l2reg,
                                      fit_intercept=True)
    for array_skl, array_fit in zip(tree_util.tree_leaves(pytree_skl),
                                    tree_util.tree_leaves(pytree_fit)):
      self.assertArraysAllClose(array_skl, array_fit, atol=1e-3)

  def test_logreg_implicit_diff(self):
    # Make sure the decorator works, evaluating the Jacobian at sklearn's sol.
    pytree_skl = test_util.logreg_skl(X=self.data[0],
                                      y=self.data[1],
                                      lam=self.default_l2reg)
    lbfgs = ScipyMinimize(fun=self.logreg_fun,
                          **self.solver_kwargs)
    def wrapper(hyperparams):
      sol_skl = pytree_skl
      return lbfgs.run(sol_skl, hyperparams, self.data).params

    jac_num = test_util.logreg_skl_jac(X=self.data[0],
                                       y=self.data[1],
                                       lam=self.default_l2reg)
    jac_custom = jax.jacrev(wrapper)(self.default_l2reg)
    for array_num, array_custom in zip(tree_util.tree_leaves(jac_num),
                                       tree_util.tree_leaves(jac_custom)):
      self.assertArraysAllClose(array_num, array_custom, atol=1e-3)

  def test_logreg_with_intercept_implicit_diff(self):
    # Make sure the decorator works, evaluating the Jacobian at sklearn's sol.
    pytree_skl = test_util.logreg_skl(X=self.data[0],
                                      y=self.data[1],
                                      lam=self.default_l2reg,
                                      fit_intercept=True)
    lbfgs = ScipyMinimize(fun=self.logreg_with_intercept_fun,
                          **self.solver_kwargs)
    def wrapper(hyperparams):
      sol_skl = {'W': pytree_skl[0], 'b': pytree_skl[1]}
      return lbfgs.run(sol_skl, hyperparams, self.data).params

    jac_num = test_util.logreg_skl_jac(X=self.data[0],
                                       y=self.data[1],
                                       lam=self.default_l2reg,
                                       fit_intercept=True)
    jac_custom = jax.jacrev(wrapper)(self.default_l2reg)
    for array_num, array_custom in zip(tree_util.tree_leaves(jac_num),
                                       tree_util.tree_leaves(jac_custom)):
      self.assertArraysAllClose(array_num, array_custom, atol=1e-3)

  def test_issue_201(self):
    # https://github.com/google/jaxopt/issues/201
    def loss(params): return params ** 2
    params = 0.0
    out = ScipyMinimize(fun=loss).run(params)  # doesn't crash
    self.assertArraysEqual(out.params, jnp.array(0.))


class ScipyBoundedMinimizeTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()

    self.n_samples, self.n_features = 50, 5
    self.data = datasets.make_regression(n_samples=self.n_samples,
                                         n_features=self.n_features,
                                         n_informative=3,
                                         random_state=0)
    self.params_init = jnp.zeros(self.n_features)
    self.bounds = (-jnp.ones(self.n_features),
                   +jnp.ones(self.n_features))
    # Formats iterates with an (arbitrary) pytree structure for testing
    # purposes.
    self.pytree_init = self._to_pytree(self.params_init)
    self.pytree_bounds = (self._to_pytree(self.bounds[0]),
                          self._to_pytree(self.bounds[1]))

    def fun(params, data):
      params = jnp.hstack([params['p0'], params['p1'], params['p2']])
      return objective.least_squares(params, data)
    self.fun = fun

    self.solver_kwargs = {'method': 'L-BFGS-B',
                          'tol': 1e-3, 'maxiter': 500}

  def _scipy_sol(self, init_params, **kwargs):
    return osp.optimize.lsq_linear(self.data[0], self.data[1], **kwargs).x

  def _scipy_box_sol(self, init_params, box_len, **kwargs):
    if 'bounds' not in kwargs:
      kwargs['bounds'] = self.bounds
    kwargs['bounds'] = (box_len * kwargs['bounds'][0],
                        box_len * kwargs['bounds'][1])
    return self._scipy_sol(init_params, **kwargs)

  def _scipy_box_jac(self, init_params, box_len, eps=1e-2, **kwargs):
    return (
        self._scipy_box_sol(init_params, box_len + eps, **kwargs) -
        self._scipy_box_sol(init_params, box_len - eps, **kwargs)) / (2. * eps)

  @staticmethod
  def _to_pytree(pytree):
    return {'p0': pytree[0], 'p1': pytree[1:3], 'p2': pytree[3:]}

  def test_fwd(self):
    lbfgs = ScipyBoundedMinimize(fun=self.fun, **self.solver_kwargs)
    pytree_fit, _ = lbfgs.run(self.pytree_init,
                              bounds=self.pytree_bounds,
                              data=self.data)

    # Checks box constraints.
    for array_lb, array_fit, array_ub in zip(
        tree_util.tree_leaves(self.pytree_bounds[0]),
        tree_util.tree_leaves(pytree_fit),
        tree_util.tree_leaves(self.pytree_bounds[1])):
      self.assertTrue(jnp.all(array_lb <= array_fit).item())
      self.assertTrue(jnp.all(array_fit <= array_ub).item())

    # Compares objective values against SciPy.
    pytree_osp = self._scipy_sol(self.params_init, bounds=self.bounds)
    pytree_osp = self._to_pytree(pytree_osp)

    for array_osp, array_fit in zip(tree_util.tree_leaves(pytree_osp),
                                    tree_util.tree_leaves(pytree_fit)):
      self.assertArraysAllClose(array_osp, array_fit, atol=1e-3)

  def test_bwd_box_len(self):
    lbfgs = ScipyBoundedMinimize(fun=self.fun,
                                 **self.solver_kwargs)
    # NOTE: cannot use solution as init since changing box_len might make the
    # init infeasible.
    def wrapper(box_len):
      scaled_bounds = (tree_scalar_mul(box_len, self.pytree_bounds[0]),
                       tree_scalar_mul(box_len, self.pytree_bounds[1]))
      return lbfgs.run(self.pytree_init, scaled_bounds, self.data).params

    box_len = 10.0
    jac_num = self._scipy_box_jac(self.params_init,
                                  box_len,
                                  bounds=self.bounds)
    jac_num = self._to_pytree(jac_num)
    jac_custom = jax.jacrev(wrapper)(box_len)
    for array_num, array_custom in zip(tree_util.tree_leaves(jac_num),
                                       tree_util.tree_leaves(jac_custom)):
      self.assertArraysAllClose(array_num, array_custom, atol=1e-2)


class ScipyRootFindingTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()

    n = 6
    key = random.PRNGKey(0)
    key_a, key_x = random.split(key)
    self.a = random.normal(key_a, [n, n])
    self.a1, self.a2 = self.a[:, :n // 2], self.a[:, n // 2:]
    self.x = random.normal(key_x, [n])
    self.x1, self.x2 = self.x[:n // 2], self.x[n // 2:]
    self.b = self.a.dot(self.x)
    self.b1, self.b2 = self.b[:n // 2], self.b[n // 2:]
    self.fun = lambda x, b: self.a.dot(x) - b
    def fun_pytree(x, b):
      o = self.a1.dot(x['x1']) + self.a2.dot(x['x2'])
      o1, o2 = o[:n // 2], o[n // 2:]
      return {'x1': o1 - b['b1'], 'x2': o2 - b['b2']}
    self.fun_pytree = fun_pytree

    self.solver_kwargs = {'method': 'hybr'}

  def test_linalg_inv(self):
    root = ScipyRootFinding(optimality_fun=self.fun, **self.solver_kwargs)
    pytree_fit, _ = root.run(jnp.zeros_like(self.x), self.b)
    self.assertArraysAllClose(pytree_fit, self.x, atol=1e-3)

  def test_linalg_inv_idf(self):
    root = ScipyRootFinding(optimality_fun=self.fun, **self.solver_kwargs)
    def wrapper(b):
      return root.run(jnp.zeros_like(self.x), b).params

    jac_theo = jnp.linalg.inv(self.a)
    jac_idf = jax.jacrev(wrapper)(self.b)
    self.assertArraysAllClose(jac_theo, jac_idf, atol=1e-3)

  @parameterized.product(pytree_type=['tuple'])
  def test_linalg_inv_pytree(self, pytree_type: str):
    pytree_init = {'x1': jnp.zeros_like(self.x1),
                   'x2': jnp.zeros_like(self.x2)}
    b = {'b1': self.b1, 'b2': self.b2}
    root = ScipyRootFinding(optimality_fun=self.fun_pytree,
                            **self.solver_kwargs)
    pytree_fit, _ = root.run(pytree_init, b)
    for array_true, array_fit in zip(tree_util.tree_leaves((self.x1, self.x2)),
                                     tree_util.tree_leaves(pytree_fit)):
      self.assertArraysAllClose(array_true, array_fit, atol=1e-3)

  @parameterized.product(pytree_type=['dict'])
  def test_linalg_inv_pytree_idf(self, pytree_type: str):
    pytree_init = {'x1': jnp.zeros_like(self.x1),
                   'x2': jnp.zeros_like(self.x2)}
    b = {'b1': self.b1, 'b2': self.b2}
    root = ScipyRootFinding(optimality_fun=self.fun_pytree,
                            **self.solver_kwargs)
    def wrapper(b):
      return root.run(pytree_init, b).params

    jac_theo = jnp.linalg.inv(self.a)
    jac_idf = jax.jacrev(wrapper)(b)
    ## `jnp.block` requires inputs to be (nested) lists.
    if pytree_type == 'dict':
      jac_idf = [list(blk_row.values()) for blk_row in jac_idf.values()]
    else:
      jac_idf = [list(blk_row) for blk_row in jac_idf]
    jac_idf = jnp.block(jac_idf)
    self.assertArraysAllClose(jac_theo, jac_idf, atol=1e-3)

  def test_broyden(self):
    # non regression test based on issue https://github.com/google/jaxopt/issues/290
    def simple_root(x):
      return x**2 - 1
    root = ScipyRootFinding(method="broyden1", optimality_fun=simple_root)
    pytree_fit, _ = root.run(jnp.zeros(5))


class ScipyLeastSquaresTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()

    self.slope = 2.5
    self.params_init = jnp.asarray([0.0, 0.0, 0.0, 0.0])
    self.params_true = jnp.asarray([1.5, -1.5, -0.25, 1.0])
    # Formats iterates with an (arbitrary) pytree structure for testing
    # purposes.
    self.pytree_init = self._to_pytree(self.params_init)
    self.pytree_true = self._to_pytree(self.params_true)

    def model_fun(params, slope, x):
      """Implements a toy non-linear curve fitting problem on 2D.

        The model is defined as
          `f(theta; x) = offset + slope * sigmoid(w1 * x1 + w2 * x2 - t0)`
        where, for the purposes of testing, `theta = (w1, w2, t0, offset)` are
        treated as the parameters to be fit via least-squares and `slope` as a
        hyperparameter used to test implicit differentiation.

      Args:
        params: a pytree containing the parameters (w1, w2, t0, offset) of the
          model.
        slope: a float with the slope hyperparameter.
        x: a np.ndarray<float>[batch, 2] representing the covariates.

      Returns:
        a np.ndarray<float>[batch] with the model output for each row in x.
      """
      w = params['w']
      t0 = params['t0']
      offset = params['offset']
      return offset + slope * jax.nn.sigmoid(jnp.dot(x, w) - t0)

    def fun(params, slope, data):
      """Computes the residuals of `model_fun` above at `data`."""
      x, y_true = data
      return y_true - model_fun(params, slope, x)
    self.fun = fun

    n_samples = 25

    key = random.PRNGKey(0)
    x = random.normal(key, [n_samples, 2])
    y_true = model_fun(self.pytree_true, self.slope, x)
    self.data = (x, y_true)
    self.onp_data = tree_util.tree_map(lambda t: onp.asarray(t, onp.float64),
                                       self.data)

    self.solver_kwargs = {'method': 'trf'}

  @staticmethod
  def _to_pytree(pytree):
    return {'w': pytree[:2], 't0': pytree[2], 'offset': pytree[3]}

  def _scipy_sol(self, init_params, slope, **kwargs):
    def scipy_fun(params):
      x, y_true = self.onp_data
      w, t0, offset = params[:2], params[2], params[3]
      y_pred = offset + slope * onp.reciprocal(1. + onp.exp(t0 - onp.dot(x, w)))
      return y_true - y_pred
    return osp.optimize.least_squares(scipy_fun, init_params, **kwargs).x

  def _scipy_slope_jac(self, init_params, slope, eps=1e-2, **kwargs):
    return (self._scipy_sol(init_params, slope + eps, **kwargs) -
            self._scipy_sol(init_params, slope - eps, **kwargs)) / (2. * eps)

  @parameterized.product(loss=['linear', 'arctan'],
                         f_scale=[0.2, 1.0])
  def test_fwd(self, loss: str, f_scale: float):
    lsq = ScipyLeastSquares(fun=self.fun,
                            loss=loss,
                            options={'f_scale': f_scale},
                            **self.solver_kwargs)
    pytree_fit, _ = lsq.run(self.pytree_init, self.slope, self.data)
    for array_true, array_fit in zip(tree_util.tree_leaves(self.pytree_true),
                                     tree_util.tree_leaves(pytree_fit)):
      self.assertArraysAllClose(array_true, array_fit, atol=1e-3)

  @parameterized.product(loss=['huber'],
                         f_scale=[0.2])
  def test_bwd(self, loss: str, f_scale: float):
    lsq = ScipyLeastSquares(fun=self.fun,
                            loss=loss,
                            options={'f_scale': f_scale},
                            **self.solver_kwargs)
    def wrapper(slope):
      return lsq.run(self.pytree_true, slope, self.data).params

    jac_num = self._scipy_slope_jac(self.params_true, self.slope,
                                    **{'loss': loss, 'f_scale': f_scale})
    jac_num = self._to_pytree(jac_num)
    jac_custom = jax.jacrev(wrapper)(self.slope)
    for array_num, array_custom in zip(tree_util.tree_leaves(jac_num),
                                       tree_util.tree_leaves(jac_custom)):
      self.assertArraysAllClose(array_num, array_custom, atol=1e-3)


class ScipyBoundedLeastSquaresTest(ScipyLeastSquaresTest):

  def setUp(self):
    super().setUp()

    lb = jnp.asarray([-1.0, -1.0, -1.0, -1.0])
    ub = jnp.asarray([+1.0, +1.0, +1.0, +1.0])
    self.bounds = (lb, ub)
    self.pytree_bounds = (self._to_pytree(lb), self._to_pytree(ub))

  def _scipy_box_sol(self, init_params, box_len, **kwargs):
    if 'bounds' not in kwargs:
      kwargs['bounds'] = self.bounds
    kwargs['bounds'] = (box_len * kwargs['bounds'][0],
                        box_len * kwargs['bounds'][1])
    return self._scipy_sol(init_params, self.slope, **kwargs)

  def _scipy_box_jac(self, init_params, box_len, eps=1e-2, **kwargs):
    return (
        self._scipy_box_sol(init_params, box_len + eps, **kwargs) -
        self._scipy_box_sol(init_params, box_len - eps, **kwargs)) / (2. * eps)

  @parameterized.product(loss=['linear', 'cauchy'],
                         f_scale=[1.0])
  def test_fwd(self, loss: str, f_scale: float):
    lsq = ScipyBoundedLeastSquares(fun=self.fun,
                                   loss=loss,
                                   options={'f_scale': f_scale},
                                   **self.solver_kwargs)
    pytree_fit, _ = lsq.run(
        self.pytree_init, self.pytree_bounds, self.slope, self.data)

    # Checks box constraints.
    for array_lb, array_fit, array_ub in zip(
        tree_util.tree_leaves(self.pytree_bounds[0]),
        tree_util.tree_leaves(pytree_fit),
        tree_util.tree_leaves(self.pytree_bounds[1])):
      self.assertTrue(jnp.all(array_lb <= array_fit).item())
      self.assertTrue(jnp.all(array_fit <= array_ub).item())

    # Compares against SciPy.
    sol_osp = self._scipy_sol(
        self.params_init,
        self.slope,
        **{'bounds': self.bounds, 'loss': loss, 'f_scale': f_scale})
    pytree_osp = self._to_pytree(sol_osp)

    for array_osp, array_fit in zip(tree_util.tree_leaves(pytree_osp),
                                    tree_util.tree_leaves(pytree_fit)):
      self.assertArraysAllClose(array_osp, array_fit, atol=1e-3)

  @parameterized.product(loss=['huber'],
                         f_scale=[1.0])
  def test_bwd_slope(self, loss: str, f_scale: float):
    sol_osp = self._scipy_sol(
        self.params_init,
        self.slope,
        **{'bounds': self.bounds, 'loss': loss, 'f_scale': f_scale})
    pytree_osp = self._to_pytree(sol_osp)

    lsq = ScipyBoundedLeastSquares(fun=self.fun,
                                   loss=loss,
                                   options={'f_scale': f_scale},
                                   **self.solver_kwargs)
    def wrapper(slope):
      return lsq.run(pytree_osp, self.pytree_bounds, slope, self.data).params

    jac_num = self._scipy_slope_jac(
        sol_osp,
        self.slope,
        **{'bounds': self.bounds, 'loss': loss, 'f_scale': f_scale})
    jac_num = self._to_pytree(jac_num)
    jac_custom = jax.jacrev(wrapper)(self.slope)
    for array_num, array_custom in zip(tree_util.tree_leaves(jac_num),
                                       tree_util.tree_leaves(jac_custom)):
      self.assertArraysAllClose(array_num, array_custom, atol=1e-2)

  @parameterized.product(loss=['soft_l1'],
                         f_scale=[0.2])
  def test_bwd_box_len(self, loss: str, f_scale: float):
    lsq = ScipyBoundedLeastSquares(fun=self.fun,
                                   loss=loss,
                                   options={'f_scale': f_scale},
                                   **self.solver_kwargs)
    # NOTE: cannot use solution as init since changing box_len might make the
    # init infeasible.
    def wrapper(box_len):
      scaled_bounds = (tree_scalar_mul(box_len, self.pytree_bounds[0]),
                       tree_scalar_mul(box_len, self.pytree_bounds[1]))
      return lsq.run(
          self.pytree_init, scaled_bounds, self.slope, self.data).params

    box_len = 1.0
    jac_num = self._scipy_box_jac(
        self.params_init,
        box_len,
        **{'bounds': self.bounds, 'loss': loss, 'f_scale': f_scale})
    jac_num = self._to_pytree(jac_num)
    jac_custom = jax.jacrev(wrapper)(box_len)
    for array_num, array_custom in zip(tree_util.tree_leaves(jac_num),
                                       tree_util.tree_leaves(jac_custom)):
      self.assertArraysAllClose(array_num, array_custom, atol=1e-2)


class BFGSInverseHessianTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()

    true_sol = [0.5, -1.2]
    # define a very simple objective
    self.objective = lambda x: (x[0] - true_sol[0])**2. + (x[1] - true_sol[1])**2.
    # analytical derivative of the objective function
    self.derivative = lambda x: [(x[0] - true_sol[0]) * 2., (x[1] - true_sol[1]) * 2.]
    # define the starting point as a random sample from the domain
    r_min, r_max = -5.0, 5.0
    onp.random.seed(6574)
    self.x0 = r_min + onp.random.rand(2) * (r_max - r_min)

  def test_inverse_hessian_bfgs(self):
    # perform the bfgs algorithm search
    res_scipy = osp.optimize.minimize(self.objective, self.x0,
                                      method='BFGS', jac=self.derivative)
    # run the same algorithm via jaxopt
    res_jaxopt = ScipyMinimize(method='BFGS', fun=self.objective).run(self.x0)
    # compare the two returned inverse Hessian matrices
    self.assertAllClose(res_scipy.hess_inv, res_jaxopt.state.hess_inv)

  def test_inverse_hessian_lbfgs(self):
    # perform the l-bfgs-b algorithm search
    res_scipy = osp.optimize.minimize(self.objective, self.x0,
                                      method='L-BFGS-B', jac=self.derivative)
    # run the same algorithm via jaxopt
    res_jaxopt = ScipyMinimize(method='L-BFGS-B', fun=self.objective).run(self.x0)
    # compare the two returned inverse Hessian matrices
    self.assertAllClose(res_scipy.hess_inv.todense(), res_jaxopt.state.hess_inv.todense())


if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



"""Model specific exceptions."""


class NotFittedError(ValueError, AttributeError):
    """Exception class to raise if estimator is used before fitting.

    This class inherits from both ValueError and AttributeError to help with
    exception handling and backward compatibility.

    Examples
    --------
    >>> from nemos.glm import GLM
    >>> from nemos.exceptions import NotFittedError
    >>> try:
    ...     GLM().predict([[[1, 2], [2, 3], [3, 4]]])
    ... except NotFittedError as e:
    ...     print(repr(e))
    NotFittedError("This GLM instance is not fitted yet. Call 'fit' with appropriate arguments.")
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax.random
import jax.numpy as jnp

import numpy as onp

from nemos.third_party.jaxopt import jaxopt
from nemos.third_party.jaxopt.jaxopt import NonlinearCG
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt._src import test_util
from sklearn import datasets

# Uncomment this line to test in x64 
# jax.config.update('jax_enable_x64', True)

def get_random_pytree():
    key = jax.random.PRNGKey(1213)

    def rn(key, l=3):
      return 0.05 * jnp.array(onp.random.normal(size=(10,)))

    def _get_random_pytree(curr_depth=0, max_depth=3):
        r = onp.random.uniform()
        if curr_depth == max_depth or r <= 0.2:  # leaf
            return rn(key)
        elif curr_depth <= 1 or r <= 0.7:  # list
            return [
                _get_random_pytree(curr_depth=curr_depth +
                                   1, max_depth=max_depth)
                for _ in range(2)
            ]
        else:  # dict
            return {
                str(_): _get_random_pytree(
                    curr_depth=curr_depth + 1, max_depth=max_depth
                )
                for _ in range(2)
            }
    return [rn(key), {'a': rn(key), 'b': rn(key)}, _get_random_pytree()]


class NonlinearCGTest(test_util.JaxoptTestCase):

  def test_arbitrary_pytree(self):
    def loss(w, data):
      X, y = data
      _w = jnp.concatenate(jax.tree_util.tree_leaves(w))
      return ((jnp.dot(X, _w) - y) ** 2).mean()

    w = get_random_pytree()
    f_w = jnp.concatenate(jax.tree_util.tree_leaves(w))
    X, y = datasets.make_classification(n_samples=15, n_features=f_w.shape[-1],
                                        n_classes=2, n_informative=3,
                                        random_state=0)
    data = (X, y)
    cg_model = NonlinearCG(fun=loss, tol=1e-2, maxiter=300,
                           method="polak-ribiere")
    w_fit, info = cg_model.run(w, data=data)
    self.assertLessEqual(info.error, 5e-2)

  @parameterized.product(
      method=["fletcher-reeves", "polak-ribiere", "hestenes-stiefel"],
      linesearch=[
          "backtracking",
          "zoom",
          jaxopt.BacktrackingLineSearch(
              objective.binary_logreg, decrease_factor=0.5
          ),
      ],
      linesearch_init=["max", "current", "increase"],
  )
  def test_binary_logreg(self, method, linesearch, linesearch_init):
    X, y = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=2, n_informative=3, random_state=0
    )
    data = (X, y)
    fun = objective.binary_logreg

    w_init = jnp.zeros(X.shape[1])
    cg_model = NonlinearCG(
        fun=fun,
        tol=1e-3,
        maxiter=100,
        method=method,
        linesearch=linesearch,
        linesearch_init=linesearch_init,
    )

    # Test with positional argument.
    w_fit, info = cg_model.run(w_init, data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, 5e-2)

    # Compare against sklearn.
    w_skl = test_util.logreg_skl(X, y, 1e-6, fit_intercept=False,
                                 multiclass=False)
    self.assertArraysAllClose(w_fit, w_skl, atol=5e-2)

  @parameterized.product(
        linesearch=['zoom', 'backtracking', 'hager-zhang'],
        method=['hestenes-stiefel', 'polak-ribiere', 'fletcher-reeves']
  )
  def test_complex(self, method, linesearch):
    """Test that optimization over complex variable z = x + jy matches equivalent real case"""

    W = jnp.array(
      [[1, - 2],
       [3, 4],
       [-4 + 2j, 5 - 3j],
       [-2 - 2j, 6]]
    )

    def C2R(z):
      return jnp.stack((z.real, z.imag))

    def R2C(x):
      return x[..., 0, :] + 1j * x[..., 1, :]

    def f(z):
      return W @ z

    def loss_complex(z):
      return jnp.sum(jnp.abs(f(z)) ** 1.5)

    def loss_real(zR):
      return loss_complex(R2C(zR))

    z0 = jnp.array([1 - 1j, 0 + 1j])
    xy0 = jnp.stack((z0.real, z0.imag))

    solver_C = NonlinearCG(fun=loss_complex, maxiter=5,
                           maxls=3, method=method, linesearch=linesearch)
    solver_R = NonlinearCG(fun=loss_real, maxiter=5,
                           maxls=3, method=method, linesearch=linesearch)
    sol_C, _ = solver_C.run(z0)
    sol_R, _ = solver_R.run(C2R(z0))
    # NOTE(vroulet): there is a slight loss of precision between real
    # and complex cases (observable for any linesearch with jax.enable_x64
    tol = 5*1e-15 if jax.config.jax_enable_x64 else 5*1e-6
    self.assertArraysAllClose(sol_C, R2C(sol_R), atol=tol, rtol=tol)


if __name__ == '__main__':
  absltest.main()



from typing import Union

import jax.numpy as jnp
import pytest
from numpy.typing import NDArray

from nemos.base_class import Base
from nemos.base_regressor import BaseRegressor


class MockBaseRegressorInvalid(BaseRegressor):
    """
    Mock model that intentionally doesn't implement all the required abstract methods.
    Used for testing the instantiation of incomplete concrete classes.
    """

    def __init__(self, std_param: int = 0):
        self.std_param = std_param
        super().__init__()

    def predict(self, X: Union[NDArray, jnp.ndarray]) -> jnp.ndarray:
        pass

    def score(self, X, y, score_type="pseudo-r2-McFadden"):
        pass

    def _get_optimal_solver_params_config(self):
        return None, None, None


class BadEstimator(Base):
    def __init__(self, param1, *args):
        super().__init__()
        pass


def test_init(mock_regressor):
    """Test the initialization of the MockBaseRegressor class."""
    assert mock_regressor.std_param == 2


def test_get_params(mock_regressor_nested):
    """Test the get_params method."""
    params = mock_regressor_nested.get_params(deep=True)
    assert params["std_param"] == 2
    assert params["other_param__std_param"] == 1


def test_set_params(mock_regressor):
    """Test the set_params method."""
    model = mock_regressor
    model.set_params(std_param=1)
    assert model.std_param == 1


def test_invalid_set_params(mock_regressor):
    """Test invalid parameter setting using the set_params method."""
    model = mock_regressor
    with pytest.raises(
        ValueError, match="Invalid parameter 'invalid_param' for estimator"
    ):
        model.set_params(invalid_param="invalid")


def test_get_param_names(mock_regressor):
    """Test retrieval of parameter names using the _get_param_names method."""
    param_names = mock_regressor._get_param_names()
    # As per your implementation, _get_param_names should capture the constructor arguments
    assert "std_param" in param_names


# To ensure abstract methods aren't callable
def test_abstract_class():
    """Ensure that abstract methods aren't callable."""
    with pytest.raises(TypeError, match="Can't instantiate abstract"):
        BaseRegressor()


def test_invalid_concrete_class():
    """Ensure that classes missing implementation of required abstract methods raise errors."""
    with pytest.raises(TypeError, match="Can't instantiate abstract"):
        MockBaseRegressorInvalid()


def test_empty_set(mock_regressor):
    """Check that an empty set_params returns self."""
    assert mock_regressor.set_params() is mock_regressor


def test_glm_varargs_error():
    """Test that variable number of argument in __init__ is not allowed."""
    bad_estimator = BadEstimator(1)
    with pytest.raises(
        RuntimeError,
        match="scikit-learn estimators should always specify their parameters",
    ):
        bad_estimator._get_param_names()



from __future__ import annotations

import copy
from functools import wraps
from typing import TYPE_CHECKING, Generator

import numpy as np

from ..typing import FeatureMatrix
from ._composition_utils import _iterate_over_components, get_input_shape, is_basis_like

if TYPE_CHECKING:
    from ._basis import Basis


def transformer_chaining(func):
    """Apply chainable methods and return self."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # Call the wrapped function and capture its return value
        result = func(*args, **kwargs)

        # If the method returns the inner `self`, replace it with the outer `self` (no deepcopy here).
        return self if result is self.basis else result

    return wrapper


class TransformerBasis:
    """Basis as ``scikit-learn`` transformers.

    This class abstracts the underlying basis function details, offering methods
    similar to scikit-learn's transformers but specifically designed for basis
    transformations. It supports fitting to data (calculating any necessary parameters
    of the basis functions), transforming data (applying the basis functions to
    data), and both fitting and transforming in one step.

    ``TransformerBasis``, unlike ``Basis``, is compatible with scikit-learn pipelining and
    model selection, enabling the cross-validation of the basis type and parameters,
    for example ``n_basis_funcs``. See the example section below.

    Parameters
    ----------
    basis :
        A concrete subclass of ``Basis``.

    Examples
    --------
    >>> from nemos.basis import BSplineEval
    >>> from nemos.basis import TransformerBasis
    >>> from nemos.glm import GLM
    >>> from sklearn.pipeline import Pipeline
    >>> from sklearn.model_selection import GridSearchCV
    >>> import numpy as np
    >>> np.random.seed(123)

    >>> # Generate data
    >>> num_samples, num_features = 10000, 1
    >>> x = np.random.normal(size=(num_samples, ))  # raw time series
    >>> basis = BSplineEval(10)
    >>> features = basis.compute_features(x)  # basis transformed time series
    >>> weights = np.random.normal(size=basis.n_basis_funcs)  # true weights
    >>> y = np.random.poisson(np.exp(features.dot(weights)))  # spike counts

    >>> # transformer can be used in pipelines
    >>> transformer = TransformerBasis(basis)
    >>> pipeline = Pipeline([ ("compute_features", transformer), ("glm", GLM()),])
    >>> pipeline = pipeline.fit(x[:, None], y)  # x need to be 2D for sklearn transformer API
    >>> out = pipeline.predict(np.arange(10)[:, None]) # predict rate from new datas
    >>> # TransformerBasis parameter can be cross-validated.
    >>> # 5-fold cross-validate the number of basis
    >>> param_grid = dict(compute_features__n_basis_funcs=[4, 10])
    >>> grid_cv = GridSearchCV(pipeline, param_grid, cv=5)
    >>> grid_cv = grid_cv.fit(x[:, None], y)
    >>> print("Cross-validated number of basis:", grid_cv.best_params_)
    Cross-validated number of basis: {'compute_features__n_basis_funcs': 10}
    """

    _chainable_methods = (
        "_set_kernel",
        "set_input_shape",
        "_set_input_independent_states",
        "setup_basis",
    )

    def __init__(self, basis: Basis):
        self._wrapped_methods = {}  # Cache for wrapped methods
        self._basis = None
        if isinstance(basis, TransformerBasis):
            basis = basis.basis
        self.basis = copy.deepcopy(basis)
        self.basis._parent = None

    @property
    def basis(self):
        """Return the wrapped basis."""
        return self._basis

    @basis.setter
    def basis(self, basis):
        if not is_basis_like(basis, sklearn_compatibility=True):
            missing_attrs = [
                attr
                for attr in ("get_params", "set_params", "compute_features")
                if not hasattr(basis, attr)
            ]
            raise TypeError(
                "TransformerBasis accepts only object implementing `get_params`, `set_params`, and `compute_features`."
                f"\nMissing methods: {missing_attrs}."
            )
        self._assign_input_shape(basis)
        self._basis = basis

    @staticmethod
    def _assign_input_shape(basis):
        # iterate over atomic or custom components
        default_shape = []
        assign_default = False
        for bas in _iterate_over_components(basis):
            ishape = get_input_shape(bas)
            assign_default |= any(i is None for i in ishape)
            ishape = [i if i is not None else () for i in ishape]
            default_shape.extend(ishape)
        if assign_default:
            basis.set_input_shape(*default_shape)
        return basis

    @staticmethod
    def _check_initialized(basis):
        if basis._input_shape_product is None:
            raise RuntimeError(
                "Cannot apply TransformerBasis: the provided basis has no defined input shape. "
                "Please call `set_input_shape` before calling `fit`, `transform`, or "
                "`fit_transform`."
            )

    def _unpack_inputs(self, X: FeatureMatrix) -> Generator:
        """Unpack inputs.

        Unpack horizontally stacked inputs using slicing. This works gracefully with ``pynapple``,
        returning a list of Tsd objects.

        Parameters
        ----------
        X:
            The inputs horizontally stacked.

        Returns
        -------
        :
            A generator looping on each individual input.

        """
        n_samples = X.shape[0]
        out = (
            np.reshape(X[:, cc : cc + n_input], (n_samples, *shape))
            for i, (bas, n_input) in enumerate(
                zip(self._iterate_over_components(), self._input_shape_product)
            )
            for shape in get_input_shape(bas)
            for cc in [sum(self._input_shape_product[:i])]
        )
        return out

    def fit(self, X: FeatureMatrix, y=None):
        """
        Check the input structure and, if necessary, compute the convolutional kernels.

        Parameters
        ----------
        X :
            The data to fit the basis functions to, shape (num_samples, num_input).
        y : ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self :
            The transformer object.

        Raises
        ------
        RuntimeError
            If ``self.n_basis_input`` is None. Call ``self.set_input_shape`` before calling ``fit`` to avoid this.
        ValueError:
            If the number of columns in X do not ``self.n_basis_input_``.

        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import MSplineEval, TransformerBasis

        >>> # Example input
        >>> X = np.random.normal(size=(100, 2))

        >>> # Define and fit tranformation basis
        >>> basis = MSplineEval(10).set_input_shape(2)
        >>> transformer = TransformerBasis(basis)
        >>> transformer_fitted = transformer.fit(X)
        """
        self._check_initialized(self.basis)
        self._check_input(X, y)
        setup_basis = getattr(self.basis, "setup_basis", None)
        if setup_basis:
            setup_basis(*self._unpack_inputs(X))
        return self

    def transform(self, X: FeatureMatrix, y=None) -> FeatureMatrix:
        """
        Transform the data using the fitted basis functions.

        Parameters
        ----------
        X :
            The data to transform using the basis functions, shape (num_samples, num_input).
        y :
            Not used, present for API consistency by convention.

        Returns
        -------
        :
            The data transformed by the basis functions.

        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import MSplineConv, TransformerBasis

        >>> # Example input
        >>> X = np.random.normal(size=(10000, 2))

        >>> basis = MSplineConv(10, window_size=200).set_input_shape(2)
        >>> transformer = TransformerBasis(basis)
        >>> # Before calling `fit` the convolution kernel is not set
        >>> transformer.kernel_

        >>> transformer_fitted = transformer.fit(X)
        >>> # Now the convolution kernel is initialized and has shape (window_size, n_basis_funcs)
        >>> transformer_fitted.kernel_.shape
        (200, 10)

        >>> # Transform basis
        >>> feature_transformed = transformer.transform(X)
        """
        self._check_initialized(self.basis)
        self._check_input(X, y)
        # transpose does not work with pynapple
        # can't use func(*X.T) to unwrap
        return self.basis.compute_features(*self._unpack_inputs(X))

    def fit_transform(self, X: FeatureMatrix, y=None) -> FeatureMatrix:
        """
        Compute the kernels and the features.

        This method is a convenience that combines fit and transform into
        one step.

        Parameters
        ----------
        X :
            The data to fit the basis functions to and then transform.
        y :
            Not used, present for API consistency by convention.

        Returns
        -------
        array-like
            The data transformed by the basis functions, after fitting the basis
            functions to the data.

        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import MSplineEval, TransformerBasis

        >>> # Example input
        >>> X = np.random.normal(size=(100, 1))

        >>> # Define tranformation basis
        >>> basis = MSplineEval(10).set_input_shape(1)
        >>> transformer = TransformerBasis(basis)

        >>> # Fit and transform basis
        >>> feature_transformed = transformer.fit_transform(X)
        """
        self.fit(X, y=y)
        return self.transform(X)

    def __getstate__(self):
        """
        Explicitly define how to pickle TransformerBasis object.

        See https://docs.python.org/3/library/pickle.html#object.__getstate__
        and https://docs.python.org/3/library/pickle.html#pickle-state
        """
        # this is the only state needed at initalization
        # returning the cached wrapped methods would create
        # a circular binding of the state to self (i.e. infinite recursion when
        # unpickling).
        return {"basis": self._basis}

    def __setstate__(self, state):
        """
        Define how to populate the object's state when unpickling.

        Note that during unpickling a new object is created without calling __init__.
        Needed to avoid infinite recursion in __getattr__ when unpickling.

        See https://docs.python.org/3/library/pickle.html#object.__setstate__
        and https://docs.python.org/3/library/pickle.html#pickle-state
        """
        self._basis = state["basis"]
        self._wrapped_methods = {}  # Reinitialize the cache

    def __getattr__(self, name: str):
        """
        Enable easy access to attributes of the underlying Basis object.

        This method caches all chainable methods (methods returning self) in a dicitonary.
        These methods are created the first time they are accessed by decorating the `self.basis.name`
        and cached for future use.

        Examples
        --------
        >>> from nemos import basis
        >>> bas = basis.RaisedCosineLinearEval(5)
        >>> trans_bas = basis.TransformerBasis(bas)
        >>> bas.n_basis_funcs
        5
        >>> trans_bas.n_basis_funcs
        5
        """
        # Check if the method has already been wrapped
        if name in self._wrapped_methods:
            return self._wrapped_methods[name]

        if not hasattr(self._basis, name) or name == "to_transformer":
            raise AttributeError(f"'TransformerBasis' object has no attribute '{name}'")

        # Get the original attribute from the basis
        attr = getattr(self._basis, name)

        # If the attribute is a callable method, wrap it dynamically
        if name in self._chainable_methods:
            wrapped = transformer_chaining(attr).__get__(self)
            self._wrapped_methods[name] = wrapped  # Cache the wrapped method
            return wrapped

        # For non-callable attributes, return them directly
        return attr

    def __setattr__(self, name: str, value) -> None:
        r"""
        Allow setting basis or the attributes of basis with a convenient dot assignment syntax.

        Setting any other attribute is not allowed.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the attribute being set is not ``basis`` or an attribute of ``basis``.

        Examples
        --------
        >>> import nemos as nmo
        >>> trans_bas = nmo.basis.TransformerBasis(nmo.basis.MSplineEval(10))
        >>> # allowed
        >>> trans_bas.basis = nmo.basis.BSplineEval(10)
        >>> # allowed
        >>> trans_bas.n_basis_funcs = 20
        >>> # not allowed
        >>> try:
        ...     trans_bas.rand_atrr = "some value"
        ... except ValueError as e:
        ...     print(repr(e))
        ValueError('Only setting basis or existing attributes of basis is allowed. Attempt to set `rand_atrr`.')
        """
        # allow self.basis = basis and other attrs of self to be retrievable
        if name in ["basis", "_wrapped_methods", "_basis"]:
            super().__setattr__(name, value)
        # allow changing existing attributes of self.basis
        elif hasattr(self.basis, name):
            setattr(self.basis, name, value)
        # don't allow setting any other attribute
        else:
            raise ValueError(
                f"Only setting basis or existing attributes of basis is allowed. Attempt to set `{name}`."
            )

    def __getitem__(self, name: str):
        """Get the basis from label and wrap into a transformer."""
        return self.__class__(self.basis.__getitem__(name))

    def __sklearn_clone__(self) -> TransformerBasis:
        """
        Customize how TransformerBasis objects are cloned when used with sklearn.model_selection.

        By default, scikit-learn tries to clone the object by calling __init__ using the output of get_params,
        which fails in our case.

        For more info: https://scikit-learn.org/stable/developers/develop.html#cloning
        """
        cloned = (
            self.basis.__sklearn_clone__()
            if hasattr(self.basis, "__sklearn_clone__")
            else copy.deepcopy(self.basis)
        )
        cloned_obj = TransformerBasis(cloned)
        return cloned_obj

    def __repr__(self):
        """Repr for Transformer wrapped basis."""
        return f"Transformer({self.basis})"

    def set_params(self, **parameters) -> TransformerBasis:
        """
        Set TransformerBasis parameters.

        When used with ``sklearn.model_selection``, users can set either the ``basis`` attribute directly
        or the parameters of the underlying Basis, but not both.

        Examples
        --------
        >>> from nemos.basis import BSplineEval, MSplineEval, TransformerBasis
        >>> basis = MSplineEval(10)
        >>> transformer_basis = TransformerBasis(basis=basis)

        >>> # setting parameters of basis is allowed
        >>> print(transformer_basis.set_params(n_basis_funcs=8).n_basis_funcs)
        8
        >>> # setting basis directly is allowed
        >>> print(type(transformer_basis.set_params(basis=BSplineEval(10)).basis))
        <class 'nemos.basis.basis.BSplineEval'>
        >>> # mixing is not allowed, this will raise an exception
        >>> try:
        ...     transformer_basis.set_params(basis=BSplineEval(10), n_basis_funcs=2)
        ... except ValueError as e:
        ...     print(repr(e))
        ValueError('Set either new basis object or parameters for existing basis, not both.')
        """
        new_basis = parameters.pop("basis", None)
        if new_basis is not None:
            self.basis = new_basis
            if len(parameters) > 0:
                raise ValueError(
                    "Set either new basis object or parameters for existing basis, not both."
                )
        else:
            self.basis = self.basis.set_params(**parameters)

        return self

    def get_params(self, deep: bool = True) -> dict:
        """Extend the dict of parameters from the underlying Basis with basis."""
        return {"basis": self.basis, **self.basis.get_params(deep)}

    def __dir__(self) -> list[str]:
        """Extend the list of properties of methods with the ones from the underlying Basis."""
        unique_attrs = set(list(super().__dir__()) + list(self.basis.__dir__()))
        # discard without raising errors if not present
        unique_attrs.discard("to_transformer")
        return list(unique_attrs)

    def __add__(self, other: TransformerBasis | Basis) -> TransformerBasis:
        """
        Add two TransformerBasis objects.

        Parameters
        ----------
        other
            The other TransformerBasis object to add.

        Returns
        -------
        : TransformerBasis
            The resulting Basis object.
        """
        add = getattr(other, "basis", other)
        return TransformerBasis(self.basis + add)

    def __rmul__(self, other: TransformerBasis | Basis | int) -> TransformerBasis:
        """Right multiplication operator for basis."""
        return self.__mul__(other)

    def __mul__(self, other: TransformerBasis | Basis | int) -> TransformerBasis:
        """
        Multiply two TransformerBasis objects.

        Parameters
        ----------
        other
            The other TransformerBasis object to multiply.

        Returns
        -------
        :
            The resulting Basis object.
        """
        mul = getattr(other, "basis", other)
        return TransformerBasis(self.basis * mul)

    def __pow__(self, exponent: int) -> TransformerBasis:
        """Exponentiation of a TransformerBasis object.

        Define the power of a basis by repeatedly applying the method __mul__.
        The exponent must be a positive integer.

        Parameters
        ----------
        exponent :
            Positive integer exponent

        Returns
        -------
        :
            The product of the basis with itself "exponent" times. Equivalent to self * self * ... * self.

        Raises
        ------
        TypeError
            If the provided exponent is not an integer.
        ValueError
            If the integer is zero or negative.
        """
        # errors are handled by Basis.__pow__
        return TransformerBasis(self.basis**exponent)

    def _check_input(self, X: FeatureMatrix, y=None):
        """Check that the input structure.

        TransformerBasis expects a 2-d array as an input. The number of columns should match the number of inputs
        the basis expects. This number can be set before the TransformerBasis is initialized, by calling
        ``Basis.set_input_shape``.

        Parameters
        ----------
        X:
            The input FeatureMatrix.

        Raises
        ------
        ValueError:
            If the input is not a 2-d array or if the number of columns does not match the expected number of inputs.
        """
        ndim = getattr(X, "ndim", None)
        if ndim is None:
            raise ValueError("The input must be a 2-dimensional array.")

        elif ndim != 2:
            raise ValueError(
                f"X must be 2-dimensional, shape (n_samples, n_features). The provided X has shape {X.shape} instead."
            )

        if X.shape[1] != sum(self._input_shape_product):
            raise ValueError(
                f"Input mismatch: expected {sum(self._input_shape_product)} inputs, but got {X.shape[1]} "
                f"columns in X.\nTo modify the required number of inputs, call `set_input_shape` before using "
                f"`fit` or `fit_transform`."
            )

        if y is not None and y.shape[0] != X.shape[0]:
            raise ValueError(
                "X and y must have the same number of samples. "
                f"X has {X.shape[0]} samples, while y has {y.shape[0]} samples."
            )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Limited-memory BFGS with box constraints."""

# This is based on:
# [1] R. H. Byrd, P. Lu and J. Nocedal. A Limited Memory Algorithm for Bound
# Constrained Optimization, (1995), SIAM Journal on Scientific and Statistical
# Computing, 16, 5, pp. 1190-1208.
# [2] J. Nocedal and S. Wright.  Numerical Optimization, second edition.

import dataclasses
import inspect
import warnings
from typing import Any, Callable, NamedTuple, Optional, Union

import jax
from jax import numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src import projection
from nemos.third_party.jaxopt.jaxopt._src.lbfgs import init_history
from nemos.third_party.jaxopt.jaxopt._src.lbfgs import update_history
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _init_stepsize
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _setup_linesearch

from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_single_dtype
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_inf_norm
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_map
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_sub
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot

import numpy as np


def _flatten_and_concat(tree: Any, batch_ndims: int = 0):
  """Flattens a pytree and concatenates leaves along the last dimension."""
  r = jax.tree_util.tree_map(
      lambda x: jnp.reshape(x, jnp.array(x).shape[:batch_ndims] + (-1,)), tree
  )
  return jax.tree_util.tree_reduce(
      lambda *args: jnp.concatenate(args, axis=-1), r
  )


def _split_and_pack_like(x: jnp.ndarray, tree: Any):
  """Splits an array and packs the components like the provided pytree."""
  treedef = jax.tree_util.tree_structure(tree)
  if jax.tree_util.treedef_is_leaf(treedef):
    return jnp.reshape(x, tree.shape)
  sizes = jax.tree_util.tree_map(jnp.size, tree)
  flat_sizes, treedef = jax.tree_util.tree_flatten(sizes)
  splits = jnp.split(x, np.cumsum(flat_sizes)[:-1])
  return jax.tree_util.tree_map(lambda y, p: jnp.reshape(y, p.shape),
                                jax.tree_util.tree_unflatten(treedef, splits),
                                tree)


def _get_error(x: Any, grad: Any, lower: Any, upper: Any):
  """Computes the error for determining convergence."""
  err = tree_map(
      lambda y, g, l, u: jnp.clip(y - g, l, u) - y, x, grad, lower, upper
  )
  return tree_inf_norm(err)


def _find_cauchy_point(
    x: jnp.ndarray,
    grad: jnp.ndarray,
    lower: jnp.ndarray,
    upper: jnp.ndarray,
    theta: jnp.ndarray,
    w: jnp.ndarray,
    m: jnp.ndarray,
):
  """Finds the Cauchy point.

  The Cauchy point is the first local minimizer of the quadratic model along the
  piecewise linear path obtained by projecting points along the steepest descent
  direction. This function implements Algorithm CP in [1].

  Args:
    x: Parameters.
    grad: Gradients with respect to parameters.
    lower: Parameter lower bounds.
    upper: Parameter upper bounds.
    theta: Scaling parameter on the identity matrix for the initial Hessian
      approximation.
    w: `W_k` matrix in equation 3.3 of [1], computed from the correction
      matrices.
    m: `M_k` matrix in equation 3.4 of [1], computed from the correction
      matrices.

  Returns:
    x_cauchy: The Cauchy point.
    c: Vector to initialize subspace minimization.
    active_set_mask: Boolean mask where `True` indicates that the coordinate is
      in the active set (not equal to the lower or upper bound).
  """
  # TODO(emilyaf, srvasude): Consider a cheaper Cauchy point approximation:
  # https://yunfei.work/lbfgsb/lbfgsb_tech_report.pdf
  eps = np.finfo(x.dtype).eps
  t = jnp.where(
      jnp.abs(grad) < eps, jnp.inf,
      jnp.where(grad < 0., (x - upper) / grad, (x - lower) / grad))
  d = jnp.where(t < eps, 0., -grad)
  x_bound = jnp.where(d > 0., upper, jnp.where(d < 0., lower, x))

  # Sort coordinates by the distance from the bounds, divided by the gradient.
  t_ind = jnp.argsort(t, axis=-1)
  t_sorted = t[t_ind]
  dt = jnp.diff(jnp.pad(t_sorted, (1, 0), "constant"))

  # Begin the loop at the first coordinate that is not at a bound.
  active_set_mask_sorted = t_sorted > eps
  start_ind = jnp.argmax(
      jnp.concatenate([active_set_mask_sorted,
                       jnp.ones([1]).astype(jnp.bool_)],
                      axis=0))

  init_c = jnp.zeros(m.shape[-1:], dtype=m.dtype)
  init_p = jnp.dot(w.T, d)
  init_df = -jnp.dot(d, d)
  init_ddf = -theta * init_df - jnp.dot(jnp.matmul(m, init_p), init_p)
  init_state = (start_ind, init_df, init_ddf, active_set_mask_sorted, init_c,
                init_p)

  def _cond(state):
    return (-state[1] / state[2] >= dt[state[0]]) & (state[0] < x.shape[-1])

  def _body(args):
    i, df, ddf, mask, c, p = args
    j = t_ind[i]  # index of the unsorted array
    c_new = c + dt[i] * p
    df_new = (
        df
        + dt[i] * ddf
        + grad[j] ** 2
        + theta * grad[j] * (x_bound[j] - x[j])
        - grad[j] * jnp.dot(w[j], jnp.matmul(m, c_new))
    )
    ddf_new = (
        ddf
        - theta * grad[j] ** 2
        - 2.0 * grad[j] * jnp.dot(w[j], jnp.matmul(m, p))
        - grad[j] ** 2 * jnp.dot(w[j], jnp.matmul(m, w[j]))
    )
    ddf_new = jnp.maximum(eps, ddf_new)
    p_new = p + grad[j] * w[j]
    return (i + 1, df_new, ddf_new, mask.at[i].set(False), c_new, p_new)

  i, df, ddf, mask_sorted, c, p = jax.lax.while_loop(_cond, _body, init_state)
  dt_min = jnp.maximum(-df / ddf, jnp.zeros([], dtype=m.dtype))
  dt_min = jnp.where(jnp.isnan(dt_min), jnp.zeros([], dtype=m.dtype), dt_min)
  t_old = (
      jax.lax.cond(
          i > 0, lambda: t_sorted[i - 1], lambda: jnp.zeros([], dtype=m.dtype)
      )
      + dt_min
  )
  active_set_mask = mask_sorted[jnp.argsort(t_ind)]
  x_cauchy = jnp.where(active_set_mask, x + t_old * d, x_bound)
  return x_cauchy, c + dt_min * p, active_set_mask


def _minimize_subspace(
    x, grad, lower, upper, x_cauchy, c, theta, w, m, active_set_mask
):
  """Direct primal method of subspace minimization from [1]."""
  w_masked = w * active_set_mask[:, jnp.newaxis].astype(w.dtype)
  r_c = (
      grad + theta * (x_cauchy - x) - jnp.dot(w_masked, jnp.matmul(m, c))
  )  # eq. 5.4

  # TODO(emilyaf): Implement the method from [1] for a large number of variables
  # and few active constraints.
  v = jnp.dot(w_masked.T, r_c)
  v = jnp.matmul(m, v)
  n = jnp.matmul(w_masked.T, w_masked) / theta
  n = jnp.eye(m.shape[-1], dtype=w.dtype) - jnp.matmul(m, n)
  v = jnp.linalg.solve(n, v)
  du = -r_c / theta - jnp.matmul(w_masked, v) / theta**2

  # TODO(emilyaf, srvasude): Investigate whether to instead truncate
  # `x_cauchy + alpha_star * du` at the boundary, following
  # https://dl.acm.org/doi/abs/10.1145/2049662.2049669.
  alpha = jnp.maximum((upper - x_cauchy) / du, (lower - x_cauchy) / du)
  alpha = jnp.where(active_set_mask & (jnp.abs(du) > 0.), alpha, 1.)
  alpha_star = jnp.min(alpha, axis=-1)
  alpha_star = jnp.minimum(alpha_star, 1.)
  return jnp.where(active_set_mask, x_cauchy + alpha_star * du, x_cauchy)


class LbfgsbState(NamedTuple):
  """Named tuple containing state information."""

  iter_num: int
  value: float
  grad: Any
  stepsize: float
  error: float
  s_history: Any
  y_history: Any
  theta: jnp.ndarray
  num_updates: jnp.ndarray
  aux: Optional[Any] = None
  failed_linesearch: bool = False

  num_fun_eval: int = 0
  num_grad_eval: int = 0
  num_linesearch_iter: int = 0


@dataclasses.dataclass(eq=False)
class LBFGSB(base.IterativeSolver):
  """L-BFGS-B solver.

  L-BFGS-B is a version of L-BFGS that incorporates box constraints on
  variables.

  Attributes:
    fun: a smooth function of the form ``fun(x, *args, **kwargs)``.
    value_and_grad: whether ``fun`` just returns the value (False) or both the
      value and gradient (True). See base.make_funs_with_aux for details.
    has_aux: whether ``fun`` outputs auxiliary data or not. If ``has_aux`` is
      False, ``fun`` is expected to be  scalar-valued. If ``has_aux`` is True,
      then we have one of the following two cases. If ``value_and_grad`` is
      False, the output should be ``value, aux = fun(...)``. If ``value_and_grad
      == True``, the output should be ``(value, aux), grad = fun(...)``. At each
      iteration of the algorithm, the auxiliary outputs are stored in
      ``state.aux``.
    maxiter: maximum number of proximal gradient descent iterations.
    tol: tolerance of the stopping criterion.
    stepsize: a stepsize to use (if <= 0, use backtracking line search), or a
      callable specifying the **positive** stepsize to use at each iteration.
    linesearch_init: strategy for line-search initialization. By default, it
      will use "increase", which will increased the step-size by a factor of
      `increase_factor` at each iteration if the step-size is larger than
      `min_stepsize`, and set it to `max_stepsize` otherwise. Other choices are
      "max", that initializes the step-size to `max_stepsize` at every
      iteration, and "current", that uses the step-size from the previous
      iteration.
    stop_if_linesearch_fails: whether to stop iterations if the line search
      fails. When True, this matches the behavior of core JAX.
    condition: Deprecated. Condition used to select the stepsize when using
      backtracking linesearch
    maxls: maximum number of iterations to use in the line search.
    decrease_factor: Deprecated. factor by which to decrease the stepsize during
      backtracking line search (default: 0.8).
    increase_factor: factor by which to increase the stepsize during line search
      (default: 1.5).
    max_stepsize: upper bound on stepsize.
    min_stepsize: lower bound on stepsize guess at start of each linesearch run.
    history_size: size of the memory to use.
    use_gamma: whether to initialize the Hessian approximation with gamma *
      theta, where gamma is chosen following equation (7.20) of 'Numerical
      Optimization' [2]. If use_gamma is set to False, theta is used as
      initialization.
    implicit_diff: whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.
    jit: whether to JIT-compile the optimization loop (default: True).
    unroll: whether to unroll the optimization loop (default: "auto").
    verbose: if set to True or 1 prints the information at each step of 
      the solver, if set to 2, print also the information of the linesearch.
  """

  fun: Callable  # pylint: disable=g-bare-generic
  value_and_grad: Union[bool, Callable] = False
  has_aux: bool = False

  maxiter: int = 50
  tol: float = 1e-3

  stepsize: Union[float, Callable[[Any], float]] = 0.0
  linesearch: str = "zoom"
  linesearch_init: str = "increase"
  stop_if_linesearch_fails: bool = False
  condition: Any = None  # deprecated in v0.8
  maxls: int = 30
  decrease_factor: Any = None  # deprecated in v0.8
  increase_factor: float = 1.5
  max_stepsize: float = 1.0
  # FIXME: should depend on whether float32 or float64 is used.
  min_stepsize: float = 1e-6

  theta: float = 1.0
  history_size: int = 10
  use_gamma: bool = True

  implicit_diff: bool = True
  implicit_diff_solve: Optional[Callable[[Any], Any]] = None

  jit: bool = True
  unroll: base.AutoOrBoolean = "auto"

  verbose: Union[bool, int] = False

  def _cond_fun(self, inputs):
    _, state = inputs[0]
    # We continue the optimization loop while the error tolerance is not met
    # and either failed linesearch is disallowed or linesearch hasn't failed.
    return (state.error > self.tol) & jnp.logical_or(
        not self.stop_if_linesearch_fails, ~state.failed_linesearch)

  def init_state(
      self,
      init_params: Any,
      bounds: Optional[Any],
      *args,
      **kwargs) -> LbfgsbState:
    """Initialize the solver state.

    Args:
      init_params: pytree containing the initial parameters.
      bounds: an optional tuple `(lb, ub)` of pytrees with structure identical
        to `init_params`, representing box constraints.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      state
    """
    if isinstance(init_params, base.OptStep):
      # `init_params` can either be a pytree or an OptStep object
      state_kwargs = dict(
          s_history=init_params.state.s_history,
          y_history=init_params.state.y_history,
          iter_num=init_params.state.iter_num,
          stepsize=init_params.state.stepsize,
          num_updates=init_params.state.num_updates,
          theta=init_params.state.theta,
      )
      init_params = init_params.params
      dtype = tree_single_dtype(init_params)
    else:
      dtype = tree_single_dtype(init_params)
      state_kwargs = dict(
          s_history=init_history(init_params, self.history_size),
          y_history=init_history(init_params, self.history_size),
          iter_num=jnp.asarray(0),
          stepsize=jnp.asarray(self.max_stepsize, dtype=dtype),
          num_updates=jnp.asarray(0),
          theta=jnp.asarray(self.theta, dtype=dtype),
      )

    (value, aux), grad = self._value_and_grad_with_aux(
        init_params, *args, **kwargs
    )
    if bounds is None:
      bounds = (tree_map(lambda x: -jnp.inf * jnp.ones_like(x), init_params),
                tree_map(lambda x: jnp.inf * jnp.ones_like(x), init_params))
    init_error = _get_error(init_params, grad, *bounds)
    return LbfgsbState(
        value=value,
        grad=grad,
        error=init_error,
        **state_kwargs,
        aux=aux,
        failed_linesearch=jnp.asarray(False),
        num_fun_eval=jnp.array(1, base.NUM_EVAL_DTYPE),
        num_grad_eval=jnp.array(1, base.NUM_EVAL_DTYPE),
        num_linesearch_iter=np.array(0, base.NUM_EVAL_DTYPE)
    )

  def update(
      self,
      params: Any,
      state: LbfgsbState,
      bounds: Optional[Any],
      *args,
      **kwargs) -> base.OptStep:
    """Performs one iteration of LBFGS.

    Args:
      params: pytree containing the parameters.
      state: named tuple containing the solver state.
      bounds: an optional tuple `(lb, ub)` of pytrees with structure identical
        to `init_params`, representing box constraints.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      (params, state)
    """
    if isinstance(params, base.OptStep):
      params = params.params
    if bounds is None:
      bounds = (tree_map(lambda x: -jnp.inf * jnp.ones_like(x), params),
                tree_map(lambda x: jnp.inf * jnp.ones_like(x), params))
    lower, upper = bounds
    x_flat = _flatten_and_concat(params)
    g_flat = _flatten_and_concat(state.grad)
    lower_flat = _flatten_and_concat(lower)
    upper_flat = _flatten_and_concat(upper)
    s_history_flat = _flatten_and_concat(state.s_history, batch_ndims=1)
    y_history_flat = _flatten_and_concat(state.y_history, batch_ndims=1)

    # Construct the limited-memory BFGS matrix from [1], section 3.
    w_flat = jnp.transpose(
        jnp.concatenate([y_history_flat, state.theta * s_history_flat], axis=0))
    s_dot_yt = jnp.matmul(s_history_flat, jnp.transpose(y_history_flat))

    # Pad extra history dimensions with constants on the diagonal to ensure
    # invertability while maintaining constant array sizes for JIT compilation.
    # The corresponding dimensions of the inverted matrix are multiplied by zero
    # downstream. If possible, the constants are chosen to equal an existing
    # nonzero element of the diagonal, for numerical stability.
    history_mask = jnp.arange(self.history_size) >= state.num_updates
    diagonal_ones = history_mask.astype(x_flat.dtype)
    i, j = jnp.diag_indices(self.history_size)
    prev_ind = state.num_updates % self.history_size - 1
    zeros = jnp.zeros_like(s_dot_yt)
    fill_diag_syt = jnp.where(state.num_updates > 0, s_dot_yt[prev_ind,
                                                              prev_ind], 1.)
    diag_adj = zeros.at[i, j].set(diagonal_ones * fill_diag_syt)

    # (m, m) tril with zeros on the diagonal
    lower_tril = jnp.tril(s_dot_yt, -1)

    # (m, m) diagonal
    diag = -1. * jnp.diag(jnp.diag(s_dot_yt)) + diag_adj

    s_dot_st = jnp.matmul(s_history_flat, jnp.transpose(s_history_flat))
    fill_diag_sst = jnp.where(state.num_updates > 0, s_dot_st[prev_ind,
                                                              prev_ind], 1.)
    sst_adj = zeros.at[i, j].set(diagonal_ones * fill_diag_sst) * state.theta
    m_inv = jnp.concatenate(  # Equation 3.4 of [1].
        [
            jnp.concatenate([diag, jnp.transpose(lower_tril)], axis=1),
            jnp.concatenate([lower_tril, sst_adj + state.theta * s_dot_st],
                            axis=1)
        ],
        axis=0)
    m = jnp.linalg.inv(m_inv)

    x_cauchy, c, active_ind = _find_cauchy_point(
        x_flat, g_flat, lower_flat, upper_flat, state.theta, w_flat, m
    )
    x_subspace_min = _minimize_subspace(
        x_flat,
        g_flat,
        lower_flat,
        upper_flat,
        x_cauchy,
        c,
        state.theta,
        w_flat,
        m,
        active_ind,
    )

    descent_direction = _split_and_pack_like(x_subspace_min - x_flat, params)

    use_linesearch = (not isinstance(self.stepsize, Callable) and
                      self.stepsize <= 0.)
    if use_linesearch:
      init_stepsize = _init_stepsize(
          self.linesearch_init,
          self.max_stepsize,
          self.min_stepsize,
          self.increase_factor,
          state.stepsize,
      )
      new_stepsize, ls_state = self.run_ls(
          init_stepsize,
          params,
          value=state.value,
          grad=state.grad,
          descent_direction=descent_direction,
          fun_args=args,
          fun_kwargs=kwargs,
      )
      new_params = ls_state.params
      new_value = ls_state.value
      new_grad = ls_state.grad
      new_aux = ls_state.aux
      failed_linesearch = ls_state.failed
      new_num_linesearch_iter = state.num_linesearch_iter + ls_state.iter_num
      new_num_grad_eval = state.num_grad_eval + ls_state.num_grad_eval
      new_num_fun_eval = state.num_fun_eval + ls_state.num_fun_eval
    else:
      if isinstance(self.stepsize, Callable):
        new_stepsize = self.stepsize(state.iter_num)
      else:
        new_stepsize = self.stepsize

      new_params = tree_add_scalar_mul(params, new_stepsize, descent_direction)
      new_params = tree_map(jnp.clip, new_params, lower, upper)

      (new_value, new_aux), new_grad = self._value_and_grad_with_aux(
          new_params, *args, **kwargs
      )
      new_num_grad_eval = state.num_grad_eval + 1
      new_num_fun_eval = state.num_fun_eval + 1
      new_num_linesearch_iter = state.num_linesearch_iter
      failed_linesearch = jnp.asarray(False)

    s = tree_sub(new_params, params)
    y = tree_sub(new_grad, state.grad)
    curvature = tree_vdot(y, s)

    if self.use_gamma:
      gamma_inv = tree_vdot(y, y) / curvature
    else:
      gamma_inv = jnp.ones([], dtype=curvature.dtype)

    history_ind = state.num_updates % self.history_size
    (new_s_history, new_y_history, new_theta, new_num_updates) = (
        jax.lax.cond(
            curvature > 0.0,
            lambda sh, yh: (  # pylint: disable=g-long-lambda
                update_history(sh, s, history_ind),
                update_history(yh, y, history_ind),
                gamma_inv * self.theta,
                state.num_updates + 1,
            ),
            lambda sh, yh: (  # pylint: disable=g-long-lambda
                sh,
                yh,
                state.theta,
                state.num_updates,
            ),
            state.s_history,
            state.y_history,
        )
    )

    error = _get_error(new_params, new_grad, lower, upper)
    new_state = LbfgsbState(
        iter_num=state.iter_num + 1,
        value=new_value,
        grad=new_grad,
        stepsize=new_stepsize,
        error=error,
        s_history=new_s_history,
        y_history=new_y_history,
        num_updates=new_num_updates,
        theta=new_theta,
        aux=new_aux,
        failed_linesearch=failed_linesearch,
        num_grad_eval=new_num_grad_eval,
        num_fun_eval=new_num_fun_eval,
        num_linesearch_iter=new_num_linesearch_iter,
    )

    if self.verbose:
      self.log_info(
          new_state,
          error_name="Projected Gradient Norm",
          additional_info={
              "Objective Value": new_value,
              "Stepsize": new_stepsize,
              "Number Linesearch Iterations": 
              new_state.num_linesearch_iter - state.num_linesearch_iter
          }
      )
    return base.OptStep(new_params, new_state)

  def _fixed_point_fun(self, sol, bounds, args, kwargs):
    step = tree_sub(sol, self._grad_fun(sol, *args, **kwargs))
    return projection.projection_box(step, bounds)

  def optimality_fun(self, sol, bounds, *args, **kwargs):
    """Optimality function mapping compatible with `@custom_root`."""
    if bounds is None:
      return self._value_and_grad_fun(sol, *args, **kwargs)[1]
    fp = self._fixed_point_fun(sol, bounds, args, kwargs)
    return tree_sub(fp, sol)

  def _value_and_grad_fun(self, params, *args, **kwargs):
    if isinstance(params, base.OptStep):
      params = params.params
    (value, _), grad = self._value_and_grad_with_aux(params, *args, **kwargs)
    return value, grad

  def _grad_fun(self, params, *args, **kwargs):
    return self._value_and_grad_fun(params, *args, **kwargs)[1]

  def __post_init__(self):
    super().__post_init__()

    _fun_with_aux, _, self._value_and_grad_with_aux = base._make_funs_with_aux(
        fun=self.fun,
        value_and_grad=self.value_and_grad,
        has_aux=self.has_aux,
    )

    # Sets up reference signature.
    fun = getattr(self.fun, "subfun", self.fun)
    signature = inspect.signature(fun)
    parameters = list(signature.parameters.values())
    new_param = inspect.Parameter(name="bounds",
                                  kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
    parameters.insert(1, new_param)
    self.reference_signature = inspect.Signature(parameters)


    unroll = self._get_unroll_option()
    linesearch_solver = _setup_linesearch(
        linesearch=self.linesearch,
        fun=_fun_with_aux,
        value_and_grad=self._value_and_grad_with_aux,
        has_aux=True,
        maxlsiter=self.maxls,
        max_stepsize=self.max_stepsize,
        jit=self.jit,
        unroll=unroll,
        verbose=max(int(self.verbose)-1, 0),
    )

    self.run_ls = linesearch_solver.run

    if self.condition is not None:
      warnings.warn("Argument condition is deprecated", DeprecationWarning)
    if self.decrease_factor is not None:
      warnings.warn(
          "Argument decrease_factor is deprecated", DeprecationWarning
      )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for L-BFGS-B."""

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp
from nemos.third_party.jaxopt.jaxopt import LBFGSB
from nemos.third_party.jaxopt.jaxopt import BacktrackingLineSearch
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import OptStep
from nemos.third_party.jaxopt.jaxopt import ScipyBoundedMinimize
from nemos.third_party.jaxopt.jaxopt._src import test_util
import numpy as onp

from sklearn import datasets

# Uncomment this line to test in x64 
# jax.config.update('jax_enable_x64', True)

class LbfgsbTest(test_util.JaxoptTestCase):

  @parameterized.parameters(
      ((0.8, 1.2), (-1.0, -2.0), (2.0, 2.0), True),
      ((2.0, 0.0), (1.5, -2.0), (3.0, 0.5), False),
      ((2.0, 0.5), (0.0, -1.0), (3.0, 1.0), "function"),
  )
  def test_correctness(self, init_params, lower, upper, value_and_grad):
    def fun(x):  # Rosenbrock function.
      return sum(100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0)
    if value_and_grad is True:
      fun = jax.value_and_grad(fun)
    elif value_and_grad == "function":
      value_and_grad = jax.value_and_grad(fun)

    x0 = jnp.array(init_params)
    bounds = (jnp.array(lower), jnp.array(upper))
    lbfgsb = LBFGSB(
        fun=fun,
        tol=1e-5,
        stepsize=-1.0,
        maxiter=100,
        history_size=5,
        use_gamma=True,
        value_and_grad=value_and_grad,
    )
    x1, _ = lbfgsb.run(x0, bounds=bounds)

    scipy_lbfgsb = ScipyBoundedMinimize(
        fun=fun,
        tol=1e-5,
        maxiter=100,
        method="L-BFGS-B",
        options={"maxcor": 5},
        value_and_grad=value_and_grad,
    )
    x2, _ = scipy_lbfgsb.run(init_params=x0, bounds=bounds)

    self.assertArraysAllClose(x1, x2, atol=5e-5)

  @parameterized.product(
      implicit_diff=[True, False],
      linesearch=["backtracking", "zoom", "hager-zhang"],
  )
  def test_Rosenbrock(self, implicit_diff, linesearch):

    def fun(x):  # Rosenbrock function.
      return 100.0 * (x["b"] - x["a"]**2.0)**2.0 + (1 - x["a"])**2.0

    x0 = {"a": jnp.zeros([]), "b": jnp.zeros([])}
    upper = {"a": jnp.asarray(2.), "b": jnp.asarray(3.)}
    lower = {"a": jnp.asarray(-1), "b": jnp.asarray(-0.5)}
    lbfgsb = LBFGSB(
        fun=fun,
        tol=1e-3,
        stepsize=-1.,
        maxiter=500,
        linesearch=linesearch,
        implicit_diff=implicit_diff)
    x, _ = lbfgsb.run(x0, bounds=(lower, upper))

    # the Rosenbrock function is zero at its minimum
    self.assertLessEqual(fun(x), 1e-3)

  @parameterized.parameters(
      ((0., -5., 0), (2., 0., 1)),
      ((0., 0., 0.), (1.5, 1., 0.4)),
  )
  def test_quadratic_bowl(self, lower, upper):
    x0 = jnp.zeros(3)
    lower = jnp.array(lower)
    upper = jnp.array(upper)
    unconstrained_optimum = jnp.array([1., -2., 0.5])
    optimum = jnp.maximum(jnp.minimum(unconstrained_optimum, upper), lower)

    f = lambda x: jnp.sum((x - unconstrained_optimum)**2)

    lbfgsb = LBFGSB(
        fun=f,
        tol=1e-3,
        maxiter=20,
        stepsize=-1.,
        linesearch="zoom")
    x, state = lbfgsb.run(x0, bounds=(lower, upper))
    self.assertLess(onp.abs(x - optimum).max(), 1e-6)

    # quadratic optimum should be found almost immediately.
    self.assertLessEqual(state.iter_num, 2)

    # Check that bounds are satisfied.
    self.assertTrue(((x >= lower) & (x <= upper)).all())

  @parameterized.product(
      use_gamma=[True, False],
      linesearch=["backtracking", "zoom", "hager-zhang"],
  )
  def test_binary_logreg(self, use_gamma, linesearch):
    x, y = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=2, n_informative=3, random_state=0
    )
    data = (x, y)
    fun = objective.binary_logreg

    w_init = jnp.zeros(x.shape[1])
    lbfgsb = LBFGSB(
        fun=fun,
        tol=1e-3,
        maxiter=500,
        use_gamma=use_gamma,
        linesearch=linesearch,
    )
    # Test with keyword argument.
    w_fit, info = lbfgsb.run(w_init, bounds=None, data=data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, 5e-2)

    # Compare against sklearn.
    w_skl = test_util.logreg_skl(
        x, y, 1e-6, fit_intercept=False, multiclass=False
    )
    self.assertArraysAllClose(w_fit, w_skl, atol=5e-2)

  @parameterized.product(
      use_gamma=[True, False],
      linesearch=[
          "backtracking",
          "zoom",
          "hager-zhang",
          BacktrackingLineSearch(objective.multiclass_logreg_with_intercept),
      ],
  )
  def test_multiclass_logreg(self, use_gamma, linesearch):
    x, y = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
    )
    data = (x, y)
    fun = objective.multiclass_logreg_with_intercept

    w_init = jnp.zeros((x.shape[1], 3))
    b_init = jnp.zeros(3)
    pytree_init = (w_init, b_init)

    u = 1e2
    l = -1e2
    upper = (u * jnp.ones_like(w_init), u * jnp.ones_like(b_init))
    lower = (l * jnp.ones_like(w_init), l * jnp.ones_like(b_init))
    lbfgsb = LBFGSB(
        fun=fun,
        tol=1e-3,
        maxiter=500,
        use_gamma=use_gamma,
        linesearch=linesearch,
    )
    # Test with positional argument.
    _, info = lbfgsb.run(pytree_init, (lower, upper), data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, 1e-2)

  def test_warm_start_hessian_approx(self):
    def fun(x):
      return sum(100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0)

    x0 = jnp.zeros(2)
    lbfgsb = LBFGSB(
        fun=fun,
        tol=1e-3,
        maxiter=2,
        stepsize=1e-3,
    )
    x1, lbfgs_state = lbfgsb.run(x0, bounds=None)

    # warm start with the previous solution
    lbfgsb.run(OptStep(x1, lbfgs_state), bounds=None)

  @parameterized.product(out_dtype=[jnp.float32, jnp.float64])
  def test_correct_dtypes(self, out_dtype):
    def f(x):
      return jnp.cos(jnp.sum(jnp.exp(-x)) ** 2).astype(out_dtype)

    with jax.experimental.enable_x64():
      x0 = jnp.ones([5, 5], dtype=jnp.float32)
      lbfgs = LBFGSB(fun=f, tol=1e-3, maxiter=500)
      x, state = lbfgs.run(x0, bounds=None)
      self.assertEqual(x.dtype, jnp.float32)
      for name in ("iter_num",):
        self.assertEqual(getattr(state, name).dtype, jnp.int64)
      for name in ("error", "s_history", "y_history", "grad"):
        self.assertEqual(getattr(state, name).dtype, jnp.float32, name)
      for name in ("value",):
        self.assertEqual(getattr(state, name).dtype, out_dtype, name)

  @parameterized.product(n_iter=[10])
  def test_n_calls(self, n_iter):
    # Test whether the number of function calls is equal to the number of
    # iterations + 1 in the no linesearch case, where the complexity is linear.
    # pylint: disable=global-variable-undefined
    global N_CALLS
    N_CALLS = 0

    def fun(x):
      global N_CALLS
      N_CALLS += 1
      return sum(100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0)

    x0 = jnp.zeros(2)
    lbfgsb = LBFGSB(
        fun=fun,
        tol=1e-7,
        maxiter=n_iter,
        stepsize=1e-3,
        jit=False,
    )
    lbfgsb.run(x0, bounds=None)

    self.assertEqual(N_CALLS, n_iter + 1)

  def test_grad_with_bounds(self):
    # Test that the gradient is correct when bounds are specified by keyword.
    # Pertinent to issue #463.
    def pipeline(x, init_pars, bounds, data):
      def fit_objective(pars, data, x):
        return -jax.scipy.stats.norm.logpdf(pars, loc=data*x, scale=1.0)
      solver = LBFGSB(fun=fit_objective, implicit_diff=True, maxiter=500, tol=1e-6)
      return solver.run(init_pars, bounds=bounds, data=data, x=x)[0]

    grad_fn = jax.grad(pipeline)
    data = jnp.array(1.5)
    res = grad_fn(0.5, jnp.array(0.0), (jnp.array(0.0), jnp.array(10.0)), data)
    self.assertEqual(res, data)
      
  def test_linear_in_box(self):
    # Fixing issue #492
    def fun(x):
      a, b = x
      return -a

    solver = LBFGSB(fun, maxiter=3)

    b = 0.
    init = jnp.array([-0.5, b])
    upper = jnp.array([1.0, 1.0])
    lower = jnp.array([-1.0, -1.0])
    result = solver.run(init, bounds=(lower, upper))
    self.assertArraysAllClose(result.params, jnp.array([1., b]))

if __name__ == "__main__":
  absltest.main()



"""
Test corner case handling in nemos.basis._composition_utils.py
"""

from contextlib import nullcontext as does_not_raise

import pytest

import nemos.basis._composition_utils as compose_utils
from nemos.basis import BSplineEval


@pytest.fixture(scope="module")
def mock_class(request):
    class Mock:
        def __init__(self, label=None):
            if label == "no-default":
                return
            elif label:
                self.label = label
            else:
                self.label = self.__class__.__name__

        def compute_features(self, x, y, *args, z=10):
            pass

    return Mock(request.param)


@pytest.fixture(scope="module")
def atomic_basis(request):
    return BSplineEval(5, label=request.param)


@pytest.mark.parametrize("mock_class", ["custom", "no-default"], indirect=True)
def test_external_class_has_default_label(mock_class):
    if hasattr(mock_class, "label"):
        assert compose_utils._has_default_label(mock_class) is None
    else:
        assert compose_utils._has_default_label(mock_class) is not None


@pytest.mark.parametrize(
    "atomic_basis, new_label, expectation",
    [
        ("label", "valid", does_not_raise()),
        ("label", 1, pytest.raises(TypeError, match="'label' must be a string")),
    ],
    indirect=["atomic_basis"],
)
def test_composition_basis_setter_label_type(atomic_basis, new_label, expectation):
    with expectation:
        exception = compose_utils._atomic_basis_label_setter_logic(
            atomic_basis, new_label
        )
        if exception:
            raise exception


@pytest.mark.parametrize("mock_class", ["custom"], indirect=True)
def test_infer_input_dimensionality(mock_class):
    assert compose_utils.infer_input_dimensionality(mock_class) == 2



"""Forward backward pass for a GLM-HMM."""

from functools import partial
from typing import Any, Callable, Optional, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
from numpy.typing import NDArray

Array = NDArray | jax.numpy.ndarray


class GLMHMMState(eqx.Module):
    """State class for the GLMHHM EM-algorithm."""

    initial_prob: Array
    transition_matrix: Array
    projection_weights: Array
    data_log_likelihood: float | Array
    iterations: int


def _analytical_m_step_initial_prob(
    posteriors: jnp.ndarray,
    is_new_session: jnp.ndarray,
    dirichlet_prior_alphas: Optional[jnp.ndarray] = None,
):
    """
    Calculate the M-step for initial state probabilities.

    Computes the maximum likelihood estimate (or MAP estimate with prior) of the
    initial state distribution by summing posterior probabilities at session starts.

    Parameters
    ----------
    posteriors :
        The posterior distribution over latent states, shape ``(n_time_bins, n_states)``.
    is_new_session :
        Boolean array indicating session start points, shape ``(n_time_bins,)``.
    dirichlet_prior_alphas :
        The parameters of the Dirichlet prior for the initial distribution,
        shape ``(n_states,)``. If None, uses a flat (uniform) prior.

    Returns
    -------
    new_initial_prob :
        Updated initial state probabilities, shape ``(n_states,)``.
        Normalized to sum to 1.
    """
    tmp_initial_prob = jnp.sum(posteriors, axis=0, where=is_new_session[:, jnp.newaxis])
    if dirichlet_prior_alphas is not None:
        tmp_initial_prob += dirichlet_prior_alphas - 1

    new_initial_prob = tmp_initial_prob / jnp.sum(tmp_initial_prob)
    return new_initial_prob


def _analytical_m_step_transition_prob(
    joint_posterior: jnp.ndarray, dirichlet_prior_alphas: Optional[jnp.ndarray] = None
):
    """
    Calculate the M-step for state transition probabilities.

    Computes the maximum likelihood estimate (or MAP estimate with prior) of the
    transition matrix by normalizing expected transition counts from the joint posterior.

    Parameters
    ----------
    joint_posterior:
        Expected counts of transitions from state i to state j,
        shape ``(n_states, n_states)``. Typically computed from the forward-backward
        algorithm as the sum over time of P(z_t=i, z_{t+1}=j | data).
    dirichlet_prior_alphas:
        The parameters of the Dirichlet prior for each row of the transition matrix,
        shape ``(n_states, n_states)``. If None, uses a flat (uniform) prior.

    Returns
    -------
    new_transition_prob:
        Updated transition probability matrix, shape ``(n_states, n_states)``.
        Each row sums to 1, where entry [i, j] is P(z_{t+1}=j | z_t=i).
    """
    if dirichlet_prior_alphas is not None:
        new_transition_prob = joint_posterior + dirichlet_prior_alphas - 1
    else:
        new_transition_prob = joint_posterior

    new_transition_prob /= jnp.sum(new_transition_prob, axis=1)[:, jnp.newaxis]
    return new_transition_prob


def compute_xi(
    alphas, betas, conditionals, normalization, is_new_session, transition_prob
):
    """
    Compute the sum of the joint posterior (xi) over consecutive latent states.

    Compute the sum of the joint posterior (xis, eqn. 13.14 of [1]_) over samples, implementing
    the summation required in the eqn. 13.19 of [1]_.

    Parameters
    ----------
    alphas :
        Forward messages, shape ``(n_time_bins, n_states)``
    betas :
        Backward messages, shape ``(n_time_bins, n_states)``.
    conditionals :
        Observation likelihoods p(y_t | z_t), shape ``(n_time_bins, n_states)``.
    normalization :
        Normalization constants from forward pass, shape ``(n_time_bins,)``.
    is_new_session :
        Boolean array, True at start of new sessions, shape ``(n_time_bins,)``.
    transition_prob :
        Transition probability matrix, shape ``(n_states, n_states)``.

    Returns
    -------
    :
        Expected joint posteriors between time steps, shape ``(n_states, n_states)``.

    References
    ----------
    .. [1] Bishop, C. M. (2006). Pattern recognition and machine learning. Springer.

    """
    # shift alpha so that alpha[t-1] aligns with beta[t]
    norm_alpha = alphas[:-1] / normalization[1:, jnp.newaxis]

    # mask out steps where t is a new session
    norm_alpha = jnp.where(is_new_session[1:, jnp.newaxis], 0.0, norm_alpha)

    # Compute xi sum in one matmul
    xi_sum = norm_alpha.T @ (conditionals[1:] * betas[1:])

    return xi_sum * transition_prob


def forward_pass(
    initial_prob: Array,
    transition_prob: Array,
    conditional_prob: Array,
    is_new_session: Array,
) -> Tuple[Array, Array]:
    """
    Forward pass of an HMM.

    This function performs the recursive forward pass over time using ``jax.lax.scan``,
    computing the filtered probabilities (``alpha``, eqn. 13.34 and 13.36 of [1]_) at each time step.
    At the start of a new session, the recursion is reset using the initial state distribution.

    Parameters
    ----------
    initial_prob :
        Initial state probability distribution, array of shape ``(n_states,)``.

    transition_prob :
        Transition matrix of shape ``(n_states, n_states)``, where entry ``T[i, j]`` is the
        probability of transitioning from state ``i`` to state ``j``.

    conditional_prob :
        Array of shape ``(n_time_bins, n_states)``, representing the observation likelihood
        ``p(y_t | z_t)`` at each time step for each state.

    is_new_session :
        Boolean array of shape ``(n_time_bins,)`` indicating the start of new sessions. When
        ``is_new_session[t]`` is True, the recursion at time ``t`` is reset using ``initial_prob``.

    Returns
    -------
    alphas :
        Array of shape ``(n_time_bins, n_states)``, containing the filtered probabilities
        at each time step. ``alphas[t]`` corresponds to the forward message at time ``t``.

    normalizers :
        Array of shape ``(n_time_bins,)`` containing the normalization constants at each
        time step. These values can be used to compute the log-likelihood of the sequence.

    Notes
    -----
    Normalization is performed at each time step to avoid numerical underflow, guarding
    against divide-by-zero errors.

    Equivalent pseudocode in standard Python:

    .. code-block:: python

        n_time_bins, n_states = py_z.shape
        alphas = np.full((n_time_bins, n_states), np.nan)
        c = np.full(n_time_bins, np.nan)

        for t in range(n_time_bins):
            if new_sess[t]:
                alphas[t] = initial_prob * py_z[t]
            else:
                alphas[t] = py_z[t] * (transition_prob.T @ alphas[t - 1])

            c[t] = np.sum(alphas[t])
            alphas[t] /= c[t]

    References
    ----------
    .. [1] Bishop, C. M. (2006). Pattern recognition and machine learning. Springer.

    """

    def initial_compute(posterior, _):
        # Equation 13.37. Reinitialize for new sessions
        return posterior * initial_prob

    def transition_compute(posterior, alpha_previous):
        # Equation 13.36
        exp_transition = jnp.matmul(transition_prob, alpha_previous)
        return posterior * exp_transition

    def body_fn(carry, xs):
        alpha_previous = carry
        posterior, is_new_session = xs
        # if it is a new session, run initial_compute
        # else, run transition_compute
        # for both functions, the inputs are posterior and alpha_previous
        alpha = jax.lax.cond(
            is_new_session,
            initial_compute,
            transition_compute,
            posterior,
            alpha_previous,
        )
        const = jnp.sum(alpha)  # Store marginal likelihood

        # Safe divide implementation so we don't divide over 0
        const = jnp.where(const > 0, const, 1.0)

        alpha = alpha / const  # Normalize - Equation 13.59
        return alpha, (alpha, const)

    init = jnp.zeros_like(conditional_prob[0])
    transition_prob = transition_prob.T
    _, (alphas, normalizers) = jax.lax.scan(
        body_fn, init, (conditional_prob, is_new_session)
    )
    return alphas, normalizers


def backward_pass(
    transition_prob: Array,
    conditional_prob: Array,
    normalizers: Array,
    is_new_session: Array,
):
    """
    Run the backward pass of the HMM inference algorithm to compute beta messages.

    This function performs the backward recursion step of the forward–backward algorithm,
    using ``jax.lax.scan`` in reverse to compute beta messages at each time step, computing
    the ``beta`` parameters, see eqn. 13.35 and 13.38 of [1]_.
    It handles session boundaries by resetting the beta messages when a new session starts.

    Parameters
    ----------
    transition_prob :
        Transition matrix of shape ``(n_states, n_states)``, where entry ``T[i, j]`` is the
        probability of transitioning from state ``i`` to state ``j``.

    conditional_prob :
        Array of shape ``(n_time_bins, n_states)``, representing the observation likelihoods
        ``p(y_t | z_t)`` at each time step for each state.

    normalizers :
        Array of shape ``(n_time_bins,)`` containing the normalization constants from the forward
        pass (e.g., sums of alpha messages). These are used to normalize the backward recursion.

    is_new_session :
        Boolean array of shape ``(n_time_bins,)`` indicating the start of new sessions. When
        ``is_new_session[t]`` is True, the backward message at time ``t`` is reset to a vector of ones.

    Returns
    -------
    betas :
        Array of shape ``(n_time_bins, n_states)``, containing the beta messages at each time step.
        The indexing is aligned with the forward pass, such that ``betas[t]`` corresponds to the
        backward message at time ``t``.

    Notes
    -----
    This implementation follows the standard HMM backward equations (Bishop, 2006, Eq. 13.38–13.39),
    including reinitialization for segmented sequences.

    Equivalent pseudocode in standard Python:

    .. code-block:: python

        n_time_bins, n_states = py_z.shape
        betas = np.full((n_time_bins, n_states), np.nan)
        betas[-1] = np.ones(n_states)

        for t in range(n_time_bins - 2, -1, -1):
            if new_sess[t + 1]:
                betas[t] = np.ones(n_states)
            else:
                betas[t] = transition_prob @ (
                        betas[t + 1] * py_z[t + 1]
                )
                betas[t] /= c[t + 1]

    References
    ----------
    .. [1] Bishop, C. M. (2006). Pattern recognition and machine learning. Springer.

    """
    init = jnp.ones_like(conditional_prob[0])

    def initial_compute(posterior, *_):
        # Initialize
        return jnp.ones_like(posterior)

    def backward_step(posterior, beta, normalization):
        # Normalize (Equation 13.62)
        return jnp.matmul(transition_prob, posterior * beta) / normalization

    def body_fn(carry, xs):
        posterior, norm, is_new_sess = xs
        beta = jax.lax.cond(
            is_new_sess,
            initial_compute,
            backward_step,
            posterior,
            carry,
            norm,
        )
        return beta, carry

    # Keeping the carrys because I am interested in
    # all outputs, including the last one.
    _, betas = jax.lax.scan(
        body_fn, init, (conditional_prob, normalizers, is_new_session), reverse=True
    )
    return betas


@partial(jax.jit, static_argnames=["inverse_link_function", "likelihood_func"])
def forward_backward(
    X: Array,
    y: Array,
    initial_prob: Array,
    transition_prob: Array,
    glm_params: Tuple[Array, Array],
    inverse_link_function: Callable,
    likelihood_func: Callable[[Array, Array], Array],
    is_new_session: Array | None = None,
):
    """
    Run the forward-backward Baum-Welch algorithm.

    Run the forward-backward Baum-Welch algorithm [1]_ that compute a posterior distribution over latent
    states. It handles session boundaries by resetting the ``alpha`` and ``beta`` messages when a new
    session starts.

    Parameters
    ----------
    X :
        Design matrix, pytree with leaves of shape ``(n_time_bins, n_features)``.

    y :
        Observations, pytree with leaves of shape ``(n_time_bins,)``.

    initial_prob :
        Initial latent state probability, pytree with leaves of shape ``(n_states, 1)``.

    transition_prob :
        Latent state transition matrix, pytree with leaves of shape ``(n_states, n_states)``.
        ``transition_prob[i, j]`` is the probability of transitioning from state ``i`` to state ``j``.

    glm_params :
        Length two tuple with the GLM coefficients of shape ``(n_features, n_states)``
        and intercept of shape ``(n_states,)``.

    inverse_link_function :
        Function mapping linear predictors to the mean of the observation distribution
        (e.g., exp for Poisson, sigmoid for Bernoulli).

    likelihood_func :
        Function computing the elementwise likelihood of observations given predicted mean values.
        Must return an array of shape ``(n_time_bins, n_states)``.

    is_new_session :
        Boolean array marking the start of a new session.
        If unspecified or empty, treats the full set of trials as a single session.

    Returns
    -------
    posteriors :
        Marginal posterior distribution over latent states, shape ``(n_time_bins, n_states)``.

    joint_posterior :
        Joint posterior distribution between consecutive time steps summed
        over samples, shape ``(n_states, n_states)``.

    log_likelihood :
        Total log-likelihood of the observation sequence under the model.

    log_likelihood_norm :
        The normalized total likelihood.

    alphas :
        Forward messages (alpha values), shape ``(n_time_bins, n_states)``.

    betas :
        Backward messages (beta values), shape ``(n_time_bins, n_states)``.

    References
    ----------
    .. [1] Bishop, C. M. (2006). *Pattern recognition and machine learning*. Springer.
    """
    coef, intercept = glm_params
    # Initialize variables
    n_time_bins = X.shape[0]

    # Revise if the data is one single session or multiple sessions.
    # If new_sess is not provided, assume one session
    if is_new_session is None:
        # default: all False, but first time bin must be True
        is_new_session = jax.lax.dynamic_update_index_in_dim(
            jnp.zeros(y.shape[0], dtype=bool), True, 0, axis=0
        )
    else:
        # use the user-provided tree, but force the first time bin to be True
        is_new_session = jax.lax.dynamic_update_index_in_dim(
            jnp.asarray(is_new_session, dtype=bool), True, 0, axis=0
        )

    # Predicted y
    if coef.ndim > 2:
        predicted_rate_given_state = inverse_link_function(
            jnp.einsum("ik, kjw->ijw", X, coef) + intercept
        )
    else:
        predicted_rate_given_state = inverse_link_function(X @ coef + intercept)

    # Compute likelihood given the fixed weights
    # Data likelihood p(y|z) from emissions model
    # NOTE:
    # For N neurons and S samples, we want the total likelihood
    # across neurons for each sample and latent state.

    # Example helper:
    # def combined_likelihood(log_likelihood_func, y, rate):
    #     # log_likelihood_func takes (y, rate) and returns log-likelihood per neuron:
    #     #   y:    shape (S, N)
    #     #   rate: shape (S, N, K)  # K = number of latent states
    #
    #     assert y.ndim == 2      # (samples, neurons)
    #     assert rate.ndim == 3   # (samples, neurons, states)
    #
    #     # vmap over the state axis: apply log_likelihood_func for each state
    #     # Result: shape (S, N, K)
    #     log_like = jax.vmap(log_likelihood_func, in_axes=(None, 2), out_axes=2)(y, rate)
    #
    #     # Combine neurons assuming conditional independence:
    #     # sum log-likelihoods over neurons (axis=1), then exponentiate for stability
    #     # Final shape: (S, K)
    #     return jnp.exp(jnp.sum(log_like, axis=1))
    #
    # Here, log_likelihood_func is the ``log_likelihood`` method from
    # nemos.observation_models.Observations with ``aggregate_sample_scores = lambda x:x``

    conditionals = likelihood_func(y, predicted_rate_given_state)

    # Compute forward pass
    alphas, normalization = forward_pass(
        initial_prob, transition_prob, conditionals, is_new_session
    )  # these are equivalent to the forward pass with python loop

    # Compute backward pass
    betas = backward_pass(transition_prob, conditionals, normalization, is_new_session)

    log_likelihood = jnp.sum(
        jnp.log(normalization)
    )  # Store log-likelihood, log of Equation 13.63

    likelihood_norm = jnp.exp(log_likelihood / n_time_bins)  # Normalize

    # Posteriors
    # ----------
    # Compute posterior distributions
    # Gamma - Equations 13.32, 13.64 from [1]
    posteriors = alphas * betas

    # xis Equations 13.43 and 13.65 from [1]
    # Posterior over consecutive states summed across time steps
    joint_posterior = compute_xi(
        alphas,
        betas,
        conditionals,
        normalization,
        is_new_session,
        transition_prob,
    )
    return (
        posteriors,
        joint_posterior,
        log_likelihood,
        likelihood_norm,
        alphas,
        betas,
    )


@partial(
    jax.jit, static_argnames=["inverse_link_function", "negative_log_likelihood_func"]
)
def hmm_negative_log_likelihood(
    glm_params: Array,
    X: Array,
    y: Array,
    posteriors: Array,
    inverse_link_function: Callable,
    negative_log_likelihood_func: Callable,
):
    """
    Compute the negative log-likelihood of the GLM-HMM.

    Compute the expected negative log-likelihood as a function of
    the projection weights. The expectation is taken over the posteriors.

    Parameters
    ----------
    glm_params:
        Projection coefficients and intercept for the GLM.
    X:
        Design matrix of observations.
    y:
        Target responses.
    posteriors:
        Posterior probabilities over states.
    inverse_link_function:
        Function mapping linear predictors to rates.
    negative_log_likelihood_func:
        Function to compute the negative log-likelihood.

    Returns
    -------
    nll:
        The scalar negative log-likelihood weighted by the posteriors.
    """
    coef, intercept = glm_params
    if coef.ndim > 2:
        # coef.shape is (n_features, n_neurons, n_states)
        predicted_rate = inverse_link_function(
            jnp.einsum("ik, kjw->ijw", X, coef) + intercept
        )
        nll = negative_log_likelihood_func(
            y,
            predicted_rate,
        ).sum(axis=1)
    else:
        predicted_rate = inverse_link_function(X @ coef + intercept)
        nll = negative_log_likelihood_func(
            y,
            predicted_rate,
        )

    # Compute dot products between log-likelihood terms and gammas
    nll = jnp.sum(nll * posteriors)

    return nll


@partial(jax.jit, static_argnames=["solver_run"])
def run_m_step(
    X: Array,
    y: Array,
    posteriors: Array,
    joint_posterior: Array,
    glm_params: Tuple[Array, Array],
    is_new_session: Array,
    solver_run: Callable[[Tuple[Array, Array], Array, Array, Array], Array],
    dirichlet_prior_alphas_init_prob: Array | None = None,
    dirichlet_prior_alphas_transition: Array | None = None,
) -> Tuple[Array, Array, Array, Any]:
    r"""
    Perform the M-step of the EM algorithm for GLM-HMM.

    Parameters
    ----------
    X:
        Design matrix of observations, shape (n_samples, n_features).
    y:
        Target responses, shape ``(n_samples,)`` or ``(n_samples, n_neurons)``.
    posteriors:
        Posterior probabilities over states, shape ``(n_samples, n_states)``.
    joint_posterior:
        Joint posterior probabilities over pairs of states summed over samples. Shape ``(n_states, n_states)``.
        :math:`\sum_t P(z_{t-1}, z_t \mid X, y, \theta_{\text{old}})`.
    glm_params:
        Current GLM coefficients and intercept terms. Coefficients have shape ``(n_features, n_states)`` for
        single observation fits and ``(n_features, n_neurons, n_states)`` for population fits. Intercepts have
        shape ``(n_states,)`` for single observation fits and ``(n_states, n_neurons)`` for population fits.
    is_new_session:
        Boolean mask marking the first observation of each session. Shape ``(n_samples,)``.
    solver_run:
        Callable performing a full optimization loop for the GLM weights.
        Note that the prior for the projection weights is baked in the solver run.
    dirichlet_prior_alphas_init_prob:
        Prior for the initial states, shape ``(n_states,)``.
    dirichlet_prior_alphas_transition:
        Prior for the transition probabilities, shape ``(n_states, n_states)``.

    Returns
    -------
    optimized_projection_weights:
        Updated projection weights after optimization.
    new_initial_prob:
        Updated initial state distribution.
    new_transition_prob:
        Updated transition matrix.
    state:
        State returned by the solver.

    Notes
    -----
    In the current implementation all Dirichlet alpha coefficients must be greater than one.
    """

    # # Update Initial state probability Eq. 13.18
    new_initial_prob = _analytical_m_step_initial_prob(
        posteriors,
        is_new_session=is_new_session,
        dirichlet_prior_alphas=dirichlet_prior_alphas_init_prob,
    )
    new_transition_prob = _analytical_m_step_transition_prob(
        joint_posterior, dirichlet_prior_alphas=dirichlet_prior_alphas_transition
    )

    # Minimize negative log-likelihood to update GLM weights
    optimized_projection_weights, state = solver_run(glm_params, X, y, posteriors)

    return optimized_projection_weights, new_initial_prob, new_transition_prob, state



"""Public available modules."""

from importlib.metadata import PackageNotFoundError as _PackageNotFoundError
from importlib.metadata import version as _get_version

from . import (
    basis,
    convolve,
    exceptions,
    fetch,
    glm,
    identifiability_constraints,
    observation_models,
    pytrees,
    regularizer,
    simulation,
    solvers,
    styles,
    tree_utils,
    type_casting,
    utils,
)
from .io.io import inspect_npz, load_model

__all__ = [
    "basis",
    "convolve",
    "exceptions",
    "fetch",
    "glm",
    "identifiability_constraints",
    "observation_models",
    "pytrees",
    "regularizer",
    "simulation",
    "styles",
    "tree_utils",
    "type_casting",
    "utils",
    "load_model",
    "inspect_npz",
]


def __dir__() -> list[str]:
    return __all__


try:
    __version__ = _get_version("nemos")
except _PackageNotFoundError:
    # package is not installed
    pass



import ast
import difflib
import itertools
import os
import pathlib
from collections import defaultdict
from typing import Dict, List, Optional

# Pairs of parameter names that are lexically similar but intentionally allowed.

# During parameter name similarity checks, some pairs of names may be flagged
# as potentially inconsistent due to their high string similarity. This list
# enumerates such known, acceptable pairs that should be *excluded* from warnings.

# Each pair is stored as a set of two strings (e.g., {"a", "a_1"}), and comparison
# is done using set equality, i.e., order does not matter.

# These typically include:
# - semantically equivalent alternatives (e.g., {"conv_time_series", "time_series"})
# - mirrored structures (e.g., {"inhib_a", "inhib_b"})
# - systematic naming conventions (e.g., {"basis1", "basis2"})
# - commonly used argument patterns (e.g., {"args", "kwargs"})
VALID_PAIRS = [
    {"conv_time_series", "time_series"},
    {"inhib_a", "inhib_b"},
    {"excit_a", "excit_b"},
    *(
        {a, b}
        for (a, b) in itertools.combinations(
            ["pytree", "pytree_1", "pytree_2", "pytree_x", "pytree_y", "pytrees"], r=2
        )
    ),
    {"args", "kwargs"},
    *({a, b} for (a, b) in itertools.combinations(["basis", "basis1", "basis2"], r=2)),
    *({a, b} for (a, b) in itertools.combinations(["axis", "axis_1", "axis_2"], r=2)),
    *(
        {a, b}
        for (a, b) in itertools.combinations(
            ["array", "array_1", "array_2", "arrays"], r=2
        )
    ),
    *(
        {a, b}
        for (a, b) in itertools.combinations(
            ["add_intercept", "intercept", "intercepts"], r=2
        )
    ),
    {"inputs", "n_inputs"},
    {"func", "funcs"},
    {"l_smooth", "l_smooth_max"},
    {"attr_name", "var_name"},
    # jaxopt solvers use fun as kwarg => our SVRG must use fun too
    # it is common for decorators to have a "func" argument, therefore I'll allow both
    {"value", "values"},
    {"l1reg", "l2reg"},
    {"bs", "bas"},
    {"func", "fun"},
    {"array", "carry"},
    {"shift", "n_shift"},
    {"state", "states"},
    {"state", "start"},
    {"states", "start"},
    {"n_states", "states"},
    {"feature_mask", "features"},
    {"observation", "observations"},
    # doc utils
    {"window_size_sec", "window_size"},
    {"glm_pos_theta", "tc_pos_theta"},
    {"predicted_firing_rate", "predicted_firing_rates"},
    {"tuning_curve", "tuning_curves"},
    {"hyperparams_prox", "hyperparams"},
    {"condition", "conditionals"},
    {"posterior", "posteriors"},
    {"beta", "betas"},
    {"alpha", "alphas"},
    {"hyperparams_prox", "hyperparams"},
    {"flat_dict", "flat_map_dict"},
    {"fit_params", "init_params"},
    {"args", "arg"},
    {"tol", "atol"},
    {"tol", "rtol"},
    {"fit_params", "flat_params"},
    {"solver_kwargs", "solver_init_kwargs"},
    {"unaccepted_name", "accepted_name"},
    {"fn", "fun"},
    {"initialize_init_proba", "initialize_transition_proba"},
    *(
        {a, b}
        for (a, b) in itertools.combinations(
            [
                "dirichlet_prior_alphas_init_prob",
                "dirichlet_prior_alphas_transition",
                "dirichlet_prior_alphas",
            ],
            r=2,
        )
    ),
]


def handle_matches(
    current_parameter: str,
    current_path: str,
    matches: List[str],
    results: Dict,
    valid_pairs: List[set[str]],
):
    """
    Handle matched parameter names by updating or creating groups in the results dictionary.

    A parameter is considered valid if it has no matches or if all its matches appear in
    `valid_pairs` as a set with the current parameter. Valid parameters are added as new entries
    in the results dictionary. Invalid parameters (i.e., those with partial or conflicting matches)
    are added to existing groups if any of their matches are already present in those groups.

    Note: This function allows overlapping groups. If `current_parameter` is similar to multiple
    parameter groups (e.g., "timin" may match both "time" and "timing"), it will be added to each
    of the matching groups independently.

    Parameters
    ----------
    current_parameter :
        The name of the parameter currently being processed.

    current_path :
        The path or context in which the parameter was found (e.g., a file or data structure path).

    matches :
        A list of other parameter names that are similar to ``current_parameter``.

    results :
        A dictionary of grouped parameters. Keys are group names, and values are dictionaries
        containing:
            - "unique_names": a set of parameter names in the group.
            - "info": a list of (parameter, path) tuples for matched entries.

    valid_pairs :
        A list of valid two-element sets. Each set contains a pair of parameter names that are
        considered equivalent or compatible.

    """
    # a parameter name is valid if no matches or all matches in valid pairs
    list_invalid = [
        match for match in matches if {match, current_parameter} not in valid_pairs
    ]
    if len(list_invalid) == 0:
        # if all matches are valid, create a new group for this parameter
        results[current_parameter] = {
            "unique_names": {current_parameter},
            "info": [(current_parameter, current_path)],
        }
    else:

        # if there is an invalid match, then add to existing result entry
        for k, v in results.items():
            # Otherwise, add the parameter to any existing groups where it has a match
            #
            # Note: We *intentionally allow overlapping groups*. If `current_parameter`
            # is similar to multiple different parameter groups
            # (e.g. "timin" may be similar to both "time" and "timing", but "time" and "timing" may
            # belong to two different groups),
            # it will be added to each of those groups.
            is_in_category = any(match in v["unique_names"] for match in list_invalid)
            if is_in_category:
                v["info"].append((current_parameter, current_path))
                v["unique_names"].add(current_parameter)


def extract_parameters_from_ast(
    tree: ast.Module,
    file_path: pathlib.Path,
    results: Dict,
    valid_pairs: List[set[str]],
    unique_param_names: set,
    similarity_cutoff: float,
):

    class ParamVisitor(ast.NodeVisitor):
        def __init__(self):
            self.class_name = None

        def visit_ClassDef(self, node):
            prev_class = self.class_name
            self.class_name = node.name
            self.generic_visit(node)
            self.class_name = prev_class

        def visit_FunctionDef(self, node):
            qualified_name = (
                f"{self.class_name}.{node.name}" if self.class_name else node.name
            )
            param_names = [str(arg.arg) for arg in node.args.args]
            for par in param_names:
                # if perfect match is present just add there
                if par in results:
                    results[par]["unique_names"].add(par)
                    results[par]["info"].append(
                        (par, f"{file_path.as_posix()}:{qualified_name}")
                    )
                    continue

                matches = difflib.get_close_matches(
                    par, unique_param_names, n=100, cutoff=similarity_cutoff
                )
                handle_matches(
                    par,
                    f"{file_path.as_posix()}:{qualified_name}",
                    matches,
                    results,
                    valid_pairs,
                )
                unique_param_names.add(par)
            self.generic_visit(node)

        def visit_AsyncFunctionDef(self, node):
            self.visit_FunctionDef(node)

    ParamVisitor().visit(tree)


def collect_similar_parameter_names_ast(
    root_dir: str | pathlib.Path,
    similarity_cutoff: float = 0.8,
    valid_pairs: Optional[List[set[str]]] = None,
) -> Dict[str, Dict]:
    if valid_pairs is None:
        valid_pairs = VALID_PAIRS

    results = {}
    unique_param_names = set()

    for dirpath, _, filenames in os.walk(root_dir):
        dirpath = pathlib.Path(dirpath)

        if "third_party" in dirpath.parts:
            continue

        for filename in filenames:
            if filename.endswith(".py"):
                full_path = dirpath / filename
                try:
                    with open(full_path, "r", encoding="utf-8") as f:
                        source = f.read()
                    tree = ast.parse(source, filename=full_path)
                    extract_parameters_from_ast(
                        tree,
                        full_path,
                        results,
                        valid_pairs,
                        unique_param_names,
                        similarity_cutoff,
                    )
                except (UnicodeDecodeError, FileNotFoundError):
                    continue

    return results


if __name__ == "__main__":
    import argparse
    import logging
    import sys

    default_path = pathlib.Path(__file__).parent.parent / "src" / "nemos"

    parser = argparse.ArgumentParser(
        description="Check parameter naming consistency using AST."
    )
    parser.add_argument(
        "--path",
        "-p",
        type=pathlib.Path,
        help="Root path to the package (source folder).",
        default=default_path,
    )
    parser.add_argument(
        "--threshold",
        "-t",
        type=float,
        default=0.8,
        help="Similarity threshold for parameter name grouping.",
    )
    args = parser.parse_args()

    logger = logging.getLogger("check_parameter_naming")
    logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")

    params = collect_similar_parameter_names_ast(
        args.path, similarity_cutoff=args.threshold
    )
    invalid = [name for name, d in params.items() if len(d["unique_names"]) > 1]

    if invalid:
        msg_lines = ["Inconsistency in parameter naming found!\n"]
        for name in invalid:
            msg_lines.append(f"{name}:\n")
            grouped_info = defaultdict(list)
            for param_name, path in sorted(params[name]["info"], key=lambda x: x[1]):
                grouped_info[param_name].append(path)
            for param_name in sorted(params[name]["unique_names"]):
                msg_lines.append(f"\t- {param_name}:\n")
                for path in grouped_info[param_name]:
                    msg_lines.append(f"\t\t- {path}\n")
            msg_lines.append("\n")
        logger.warning("".join(msg_lines))
        sys.exit(1)
    else:
        logger.info("No parameter naming inconsistencies found.")



import warnings

import jax
import jax.numpy as jnp
import numpy as np
import pytest

import nemos as nmo
from nemos.glm.initialize_parameters import initialize_intercept_matching_mean_rate


@pytest.mark.parametrize(
    "non_linearity",
    [
        jnp.exp,
        jax.nn.softplus,
        lambda x: jnp.exp(x),
        jax.nn.sigmoid,
        jax.lax.logistic,
        lambda x: jax.lax.logistic(x),
        jax.scipy.special.expit,
        jax.scipy.stats.norm.cdf,
    ],
)
@pytest.mark.parametrize(
    "output_y",
    [np.random.uniform(0, 1, size=(10,)), np.random.uniform(0, 1, size=(10, 2))],
)
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_invert_non_linearity(non_linearity, output_y):
    inv_y = initialize_intercept_matching_mean_rate(
        inverse_link_function=non_linearity, y=output_y
    )
    assert jnp.allclose(non_linearity(inv_y), jnp.mean(output_y, axis=0), rtol=10**-5)


@pytest.mark.parametrize(
    "non_linearity, expectation",
    [
        (jnp.exp, pytest.raises(ValueError, match=".+The mean firing rate has")),
        (
            jax.nn.softplus,
            pytest.raises(ValueError, match=".+The mean firing rate has"),
        ),
        (
            lambda x: jnp.exp(x),
            pytest.raises(
                ValueError, match=".+Please, provide initial parameters instead"
            ),
        ),
        (
            jax.nn.sigmoid,
            pytest.raises(
                ValueError, match=".+Please, provide initial parameters instead"
            ),
        ),
        (
            jax.lax.logistic,
            pytest.raises(ValueError, match=".+The mean firing rate has"),
        ),
        (
            lambda x: jax.lax.logistic(x),
            pytest.raises(
                ValueError, match=".+Please, provide initial parameters instead"
            ),
        ),
        (
            jax.scipy.stats.norm.cdf,
            pytest.raises(ValueError, match=".+The mean firing rate has"),
        ),
    ],
)
def test_initialization_error_nan_input(non_linearity, expectation):
    """Initialize invalid."""
    output_y = np.full((10, 2), np.nan)
    with expectation:
        initialize_intercept_matching_mean_rate(
            inverse_link_function=non_linearity, y=output_y
        )


def test_initialization_error_non_invertible():
    """Initialize invalid."""
    output_y = np.random.uniform(size=100)
    inv_link = lambda x: jax.nn.softplus(x) + 10
    with pytest.raises(
        ValueError, match="Failed to initialize the model intercept.+Please, provide"
    ):
        with warnings.catch_warnings():
            # ignore the warning raised by the root-finder (there is no root)
            warnings.filterwarnings(
                "ignore", category=RuntimeWarning, message="Tolerance of"
            )
            initialize_intercept_matching_mean_rate(
                inverse_link_function=inv_link, y=output_y
            )
